# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

"""This module imports some commonly used methods and classes.

You can use ``from pymor.basic import *`` in interactive session
to have the most important parts of pyMOR directly available.
"""

# flake8: noqa

from pymor.algorithms.basic import almost_equal, relative_error, project_array
from pymor.algorithms.ei import interpolate_operators, interpolate_function, ei_greedy, deim
from pymor.algorithms.error import reduction_error_analysis
from pymor.algorithms.gram_schmidt import gram_schmidt, gram_schmidt_biorth
from pymor.algorithms.greedy import rb_greedy
from pymor.algorithms.adaptivegreedy import rb_adaptive_greedy
from pymor.algorithms.newton import newton
from pymor.algorithms.pod import pod
from pymor.algorithms.preassemble import preassemble
from pymor.algorithms.projection import project, project_to_subbasis
from pymor.algorithms.simplify import expand

from pymor.analyticalproblems.burgers import burgers_problem, burgers_problem_2d
from pymor.analyticalproblems.domaindescriptions import (RectDomain, CylindricalDomain, TorusDomain, LineDomain,
                                                         CircleDomain)
from pymor.analyticalproblems.domaindescriptions import DiscDomain, CircularSectorDomain, PolygonalDomain
from pymor.analyticalproblems.elliptic import StationaryProblem
from pymor.analyticalproblems.functions import (ConstantFunction, GenericFunction, ExpressionFunction, LincombFunction,
                                                BitmapFunction)
from pymor.analyticalproblems.helmholtz import helmholtz_problem
from pymor.analyticalproblems.instationary import InstationaryProblem
from pymor.analyticalproblems.thermalblock import thermal_block_problem
from pymor.analyticalproblems.text import text_problem

from pymor.core.cache import clear_caches, enable_caching, disable_caching
from pymor.core.defaults import print_defaults, write_defaults_to_file, load_defaults_from_file, set_defaults
from pymor.core.logger import set_log_levels, getLogger
from pymor.core.pickle import dump, dumps, load, loads

from pymor.models.basic import StationaryModel, InstationaryModel
from pymor.models.iosys import LTIModel, SecondOrderModel, TransferFunction

from pymor.discretizers.builtin import (discretize_stationary_cg, discretize_instationary_cg,
                                        discretize_stationary_fv, discretize_instationary_fv,
                                        OnedGrid, TriaGrid, RectGrid, load_gmsh)
from pymor.discretizers.builtin.domaindiscretizers.default import discretize_domain_default
from pymor.discretizers.builtin.grids.boundaryinfos import (EmptyBoundaryInfo, GenericBoundaryInfo,
                                                            AllDirichletBoundaryInfo)

from pymor.operators.constructions import (LincombOperator, ConcatenationOperator,
                                           ComponentProjectionOperator, IdentityOperator, ConstantOperator,
                                           ZeroOperator, VectorArrayOperator, VectorOperator, VectorFunctional,
                                           FixedParameterOperator, AdjointOperator, SelectionOperator, induced_norm)
from pymor.operators.ei import EmpiricalInterpolatedOperator
from pymor.operators.numpy import NumpyGenericOperator, NumpyMatrixOperator

from pymor.parallel.default import new_parallel_pool
from pymor.parallel.manager import RemoteObjectManager

from pymor.parameters.base import Parameters, Mu, ParametricObject, ParameterSpace
from pymor.parameters.functionals import (ProjectionParameterFunctional, GenericParameterFunctional,
                                          ExpressionParameterFunctional)

from pymor.reductors.basic import StationaryRBReductor, InstationaryRBReductor, LTIPGReductor, SOLTIPGReductor
from pymor.reductors.bt import BTReductor, LQGBTReductor, BRBTReductor
from pymor.reductors.coercive import CoerciveRBReductor, SimpleCoerciveRBReductor
from pymor.reductors.h2 import IRKAReductor, OneSidedIRKAReductor, TSIAReductor, TFIRKAReductor
from pymor.reductors.interpolation import LTIBHIReductor, SOBHIReductor, TFBHIReductor
from pymor.reductors.mt import MTReductor
from pymor.reductors.parabolic import ParabolicRBReductor
from pymor.reductors.sobt import (SOBTpReductor, SOBTvReductor, SOBTpvReductor, SOBTvpReductor, SOBTfvReductor,
                                  SOBTReductor)
from pymor.reductors.sor_irka import SORIRKAReductor

from pymor.tools.random import default_random_state

from pymor.vectorarrays.constructions import cat_arrays
from pymor.vectorarrays.list import ListVectorSpace
from pymor.vectorarrays.numpy import NumpyVectorSpace



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

"""|Operators| based on |NumPy| arrays.

This module provides the following |NumPy|-based |Operators|:

- |NumpyMatrixOperator| wraps a 2D |NumPy array| as an |Operator|.
- |NumpyMatrixBasedOperator| should be used as base class for all |Operators|
  which assemble into a |NumpyMatrixOperator|.
- |NumpyGenericOperator| wraps an arbitrary Python function between
  |NumPy arrays| as an |Operator|.
"""

from functools import reduce

import numpy as np
from scipy.io import mmwrite, savemat
from scipy.linalg import solve
import scipy.sparse
from scipy.sparse import issparse

from pymor.core.base import abstractmethod
from pymor.core.config import config
from pymor.core.defaults import defaults
from pymor.core.exceptions import InversionError
from pymor.core.logger import getLogger
from pymor.operators.interface import Operator
from pymor.vectorarrays.numpy import NumpyVectorSpace


class NumpyGenericOperator(Operator):
    """Wraps an arbitrary Python function between |NumPy arrays| as an |Operator|.

    Parameters
    ----------
    mapping
        The function to wrap. If `parameters` is `None`, the function is of
        the form `mapping(U)` and is expected to be vectorized. In particular::

            mapping(U).shape == U.shape[:-1] + (dim_range,).

        If `parameters` is not `None`, the function has to have the signature
        `mapping(U, mu)`.
    adjoint_mapping
        The adjoint function to wrap. If `parameters` is `None`, the function is of
        the form `adjoint_mapping(U)` and is expected to be vectorized. In particular::

            adjoint_mapping(U).shape == U.shape[:-1] + (dim_source,).

        If `parameters` is not `None`, the function has to have the signature
        `adjoint_mapping(U, mu)`.
    dim_source
        Dimension of the operator's source.
    dim_range
        Dimension of the operator's range.
    linear
        Set to `True` if the provided `mapping` and `adjoint_mapping` are linear.
    parameters
        The |Parameters| the depends on.
    solver_options
        The |solver_options| for the operator.
    name
        Name of the operator.
    """

    def __init__(self, mapping, adjoint_mapping=None, dim_source=1, dim_range=1, linear=False, parameters={},
                 source_id=None, range_id=None, solver_options=None, name=None):
        self.__auto_init(locals())
        self.source = NumpyVectorSpace(dim_source, source_id)
        self.range = NumpyVectorSpace(dim_range, range_id)
        self.parameters_own = parameters

    def apply(self, U, mu=None):
        assert U in self.source
        assert self.parameters.assert_compatible(mu)
        if self.parametric:
            return self.range.make_array(self.mapping(U.to_numpy(), mu=mu))
        else:
            return self.range.make_array(self.mapping(U.to_numpy()))

    def apply_adjoint(self, V, mu=None):
        if self.adjoint_mapping is None:
            raise ValueError('NumpyGenericOperator: adjoint mapping was not defined.')
        assert V in self.range
        assert self.parameters.assert_compatible(mu)
        V = V.to_numpy()
        if self.parametric:
            return self.source.make_array(self.adjoint_mapping(V, mu=mu))
        else:
            return self.source.make_array(self.adjoint_mapping(V))


class NumpyMatrixBasedOperator(Operator):
    """Base class for operators which assemble into a |NumpyMatrixOperator|.

    Attributes
    ----------
    sparse
        `True` if the operator assembles into a sparse matrix, `False` if the
        operator assembles into a dense matrix, `None` if unknown.
    """

    linear = True
    sparse = None

    @property
    def H(self):
        if not self.parametric:
            return self.assemble().H
        else:
            return super().H

    @abstractmethod
    def _assemble(self, mu=None):
        pass

    def assemble(self, mu=None):
        assert self.parameters.assert_compatible(mu)
        return NumpyMatrixOperator(self._assemble(mu),
                                   source_id=self.source.id,
                                   range_id=self.range.id,
                                   solver_options=self.solver_options,
                                   name=self.name)

    def apply(self, U, mu=None):
        return self.assemble(mu).apply(U)

    def apply_adjoint(self, V, mu=None):
        return self.assemble(mu).apply_adjoint(V)

    def as_range_array(self, mu=None):
        return self.assemble(mu).as_range_array()

    def as_source_array(self, mu=None):
        return self.assemble(mu).as_source_array()

    def apply_inverse(self, V, mu=None, initial_guess=None, least_squares=False):
        return self.assemble(mu).apply_inverse(V, initial_guess=initial_guess, least_squares=least_squares)

    def export_matrix(self, filename, matrix_name=None, output_format='matlab', mu=None):
        """Save the matrix of the operator to a file.

        Parameters
        ----------
        filename
            Name of output file.
        matrix_name
            The name, the output matrix is given. (Comment field is used in
            case of Matrix Market output_format.) If `None`, the |Operator|'s `name`
            is used.
        output_format
            Output file format. Either `matlab` or `matrixmarket`.
        mu
            The |parameter values| to assemble the to be exported matrix for.
        """
        assert output_format in {'matlab', 'matrixmarket'}
        matrix = self.assemble(mu).matrix
        matrix_name = matrix_name or self.name
        if output_format == 'matlab':
            savemat(filename, {matrix_name: matrix})
        else:
            mmwrite(filename, matrix, comment=matrix_name)


class NumpyMatrixOperator(NumpyMatrixBasedOperator):
    """Wraps a 2D |NumPy Array| as an |Operator|.

    Parameters
    ----------
    matrix
        The |NumPy array| which is to be wrapped.
    source_id
        The id of the operator's `source` |VectorSpace|.
    range_id
        The id of the operator's `range` |VectorSpace|.
    solver_options
        The |solver_options| for the operator.
    name
        Name of the operator.
    """

    def __init__(self, matrix, source_id=None, range_id=None, solver_options=None, name=None):
        assert matrix.ndim <= 2
        if matrix.ndim == 1:
            matrix = np.reshape(matrix, (1, -1))
        try:
            matrix.setflags(write=False)  # make numpy arrays read-only
        except AttributeError:
            pass

        self.__auto_init(locals())
        self.source = NumpyVectorSpace(matrix.shape[1], source_id)
        self.range = NumpyVectorSpace(matrix.shape[0], range_id)
        self.sparse = issparse(matrix)

    @classmethod
    def from_file(cls, path, key=None, source_id=None, range_id=None, solver_options=None, name=None):
        from pymor.tools.io import load_matrix
        matrix = load_matrix(path, key=key)
        return cls(matrix, solver_options=solver_options, source_id=source_id, range_id=range_id,
                   name=name or key or path)

    @property
    def H(self):
        options = {'inverse': self.solver_options.get('inverse_adjoint'),
                   'inverse_adjoint': self.solver_options.get('inverse')} if self.solver_options else None
        if self.sparse:
            adjoint_matrix = self.matrix.transpose(copy=False).conj(copy=False)
        elif np.isrealobj(self.matrix):
            adjoint_matrix = self.matrix.T
        else:
            adjoint_matrix = self.matrix.T.conj()
        return self.with_(matrix=adjoint_matrix, source_id=self.range_id, range_id=self.source_id,
                          solver_options=options, name=self.name + '_adjoint')

    def _assemble(self, mu=None):
        pass

    def assemble(self, mu=None):
        return self

    def as_range_array(self, mu=None):
        if self.sparse:
            return Operator.as_range_array(self)
        return self.range.from_numpy(self.matrix.T.copy())

    def as_source_array(self, mu=None):
        if self.sparse:
            return Operator.as_source_array(self)
        return self.source.from_numpy(self.matrix.copy()).conj()

    def apply(self, U, mu=None):
        assert U in self.source
        return self.range.make_array(self.matrix.dot(U.to_numpy().T).T)

    def apply_adjoint(self, V, mu=None):
        assert V in self.range
        return self.H.apply(V, mu=mu)

    @defaults('check_finite', 'default_sparse_solver_backend')
    def apply_inverse(self, V, mu=None, initial_guess=None, least_squares=False,
                      check_finite=True, default_sparse_solver_backend='scipy'):
        """Apply the inverse operator.

        Parameters
        ----------
        V
            |VectorArray| of vectors to which the inverse operator is applied.
        mu
            The |parameter values| for which to evaluate the inverse operator.
        initial_guess
            |VectorArray| with the same length as `V` containing initial guesses
            for the solution.  Some implementations of `apply_inverse` may
            ignore this parameter.  If `None` a solver-dependent default is used.
        least_squares
            If `True`, solve the least squares problem::

                u = argmin ||op(u) - v||_2.

            Since for an invertible operator the least squares solution agrees
            with the result of the application of the inverse operator,
            setting this option should, in general, have no effect on the result
            for those operators. However, note that when no appropriate
            |solver_options| are set for the operator, most implementations
            will choose a least squares solver by default which may be
            undesirable.
        check_finite
            Test if solution only contains finite values.
        default_sparse_solver_backend
            Default sparse solver backend to use (scipy, pyamg, generic).

        Returns
        -------
        |VectorArray| of the inverse operator evaluations.

        Raises
        ------
        InversionError
            The operator could not be inverted.
        """
        assert V in self.range
        assert initial_guess is None or initial_guess in self.source and len(initial_guess) == len(V)

        if V.dim == 0:
            if self.source.dim == 0 or least_squares:
                return self.source.make_array(np.zeros((len(V), self.source.dim)))
            else:
                raise InversionError

        if self.source.dim != self.range.dim and not least_squares:
            raise InversionError

        options = self.solver_options.get('inverse') if self.solver_options else None
        assert self.sparse or not options

        if self.sparse:
            if options:
                solver = options if isinstance(options, str) else options['type']
                backend = solver.split('_')[0]
            else:
                backend = default_sparse_solver_backend

            if backend == 'scipy':
                from pymor.bindings.scipy import apply_inverse as apply_inverse_impl
            elif backend == 'pyamg':
                if not config.HAVE_PYAMG:
                    raise RuntimeError('PyAMG support not enabled.')
                from pymor.bindings.pyamg import apply_inverse as apply_inverse_impl
            elif backend == 'generic':
                logger = getLogger('pymor.bindings.scipy.scipy_apply_inverse')
                logger.warning('You have selected a (potentially slow) generic solver for a NumPy matrix operator!')
                from pymor.algorithms.genericsolvers import apply_inverse as apply_inverse_impl
            else:
                raise NotImplementedError

            return apply_inverse_impl(self, V, initial_guess=initial_guess, options=options,
                                      least_squares=least_squares, check_finite=check_finite)

        else:
            if least_squares:
                try:
                    R, _, _, _ = np.linalg.lstsq(self.matrix, V.to_numpy().T, rcond=None)
                except np.linalg.LinAlgError as e:
                    raise InversionError(f'{str(type(e))}: {str(e)}')
                R = R.T
            else:
                try:
                    R = solve(self.matrix, V.to_numpy().T).T
                except np.linalg.LinAlgError as e:
                    raise InversionError(f'{str(type(e))}: {str(e)}')

            if check_finite:
                if not np.isfinite(np.sum(R)):
                    raise InversionError('Result contains non-finite values')

            return self.source.make_array(R)

    def apply_inverse_adjoint(self, U, mu=None, initial_guess=None, least_squares=False):
        return self.H.apply_inverse(U, mu=mu, initial_guess=initial_guess, least_squares=least_squares)

    def _assemble_lincomb(self, operators, coefficients, identity_shift=0., solver_options=None, name=None):
        if not all(isinstance(op, NumpyMatrixOperator) for op in operators):
            return None

        common_mat_dtype = reduce(np.promote_types,
                                  (op.matrix.dtype for op in operators if hasattr(op, 'matrix')))
        common_coef_dtype = reduce(np.promote_types, (type(c) for c in coefficients + [identity_shift]))
        common_dtype = np.promote_types(common_mat_dtype, common_coef_dtype)

        if coefficients[0] == 1:
            matrix = operators[0].matrix.astype(common_dtype)
        else:
            matrix = operators[0].matrix * coefficients[0]
            if matrix.dtype != common_dtype:
                matrix = matrix.astype(common_dtype)

        for op, c in zip(operators[1:], coefficients[1:]):
            if c == 1:
                try:
                    matrix += op.matrix
                except NotImplementedError:
                    matrix = matrix + op.matrix
            elif c == -1:
                try:
                    matrix -= op.matrix
                except NotImplementedError:
                    matrix = matrix - op.matrix
            else:
                try:
                    matrix += (op.matrix * c)
                except NotImplementedError:
                    matrix = matrix + (op.matrix * c)

        if identity_shift != 0:
            if identity_shift.imag == 0:
                identity_shift = identity_shift.real
            if operators[0].sparse:
                try:
                    matrix += (scipy.sparse.eye(matrix.shape[0]) * identity_shift)
                except NotImplementedError:
                    matrix = matrix + (scipy.sparse.eye(matrix.shape[0]) * identity_shift)
            else:
                matrix += (np.eye(matrix.shape[0]) * identity_shift)

        return NumpyMatrixOperator(matrix,
                                   source_id=self.source.id,
                                   range_id=self.range.id,
                                   solver_options=solver_options)

    def __getstate__(self):
        if hasattr(self.matrix, 'factorization'):  # remove unplicklable SuperLU factorization
            del self.matrix.factorization
        return self.__dict__

    def _format_repr(self, max_width, verbosity):
        if self.sparse:
            matrix_repr = f'<{self.range.dim}x{self.source.dim} sparse, {self.matrix.nnz} nnz>'
        else:
            matrix_repr = f'<{self.range.dim}x{self.source.dim} dense>'
        return super()._format_repr(max_width, verbosity, override={'matrix': matrix_repr})



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

from pymor.core.config import config
from pymor.core.exceptions import IOLibsMissing

if config.HAVE_VTKIO:

    from pymor.discretizers.builtin.grids.io import to_meshio
    from pymor.tools.io.vtk import write_vtk_collection

    def write_vtk(grid, data, filename_base, codim=2, metadata=None):
        """Convenience function around write_vtk_collection(to_meshio(...))"""
        return write_vtk_collection(meshes=to_meshio(grid, data, codim=codim),
                                    filename_base=filename_base, metadata=metadata)

else:

    def write_vtk(*args, **kwargs):
        raise IOLibsMissing()



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

"""This module provides plotting support inside the Jupyter notebook.

To use these routines you first have to execute ::

        %matplotlib notebook

inside the given notebook.
"""
import IPython

from pymor.core.defaults import defaults
from pymor.discretizers.builtin.gui.jupyter.matplotlib import visualize_patch
from pymor.core.config import config

# AFAICT there is no robust way to query for loaded extensions
# and we have to make sure we do not setup two redirects
_extension_loaded = False

@defaults('backend')
def get_visualizer(backend='py3js'):
    if backend not in ('py3js', 'MPL'):
        raise ValueError
    if backend == 'py3js' and config.HAVE_PYTHREEJS:
        from pymor.discretizers.builtin.gui.jupyter.threejs import visualize_py3js
        return visualize_py3js
    else:
        return visualize_patch


def progress_bar(sequence, every=None, size=None, name='Parameters'):
    from ipywidgets import IntProgress, HTML, VBox
    # c&p from https://github.com/kuk/log-progress
    is_iterator = False
    if size is None:
        try:
            size = len(sequence)
        except TypeError:
            is_iterator = True
    if size is not None:
        if every is None:
            if size <= 200:
                every = 1
            else:
                every = int(size / 200)     # every 0.5%
    else:
        assert every is not None, 'sequence is iterator, set every'

    if is_iterator:
        progress = IntProgress(min=0, max=1, value=1)
        progress.bar_style = 'info'
    else:
        progress = IntProgress(min=0, max=size, value=0)
    label = HTML()
    box = VBox(children=[label, progress])
    IPython.display.display(box)

    index = 0
    try:
        for index, record in enumerate(sequence, 1):
            if index == 1 or index % every == 0:
                if is_iterator:
                    label.value = '{name}: {index} / ?'.format(
                        name=name,
                        index=index
                    )
                else:
                    progress.value = index
                    label.value = u'{name}: {index} / {size}'.format(
                        name=name,
                        index=index,
                        size=size
                    )
            yield record
    except:
        progress.bar_style = 'danger'
        raise
    else:
        progress.bar_style = 'success'
        progress.value = index
        label.value = f"{name}: {str(index or '?')}"


def load_ipython_extension(ipython):
    global _extension_loaded
    from pymor.discretizers.builtin.gui.jupyter.logging import redirect_logging
    ipython.events.register('pre_run_cell', redirect_logging.start)
    ipython.events.register('post_run_cell', redirect_logging.stop)
    _extension_loaded = True


def unload_ipython_extension(ipython):
    global _extension_loaded
    from pymor.discretizers.builtin.gui.jupyter.logging import redirect_logging
    ipython.events.unregister('pre_run_cell', redirect_logging.start)
    ipython.events.unregister('post_run_cell', redirect_logging.stop)
    _extension_loaded = False



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

from numbers import Number

import numpy as np

from pymor.core.base import BasicObject, abstractmethod, abstractclassmethod, classinstancemethod
from pymor.tools.random import get_random_state
from pymor.vectorarrays.interface import VectorArray, VectorSpace, _create_random_values


class Vector(BasicObject):
    """Interface for vectors used in conjunction with |ListVectorArray|.

    This interface must be satisfied by the individual entries of the
    vector `list` managed by |ListVectorArray|. All interface methods
    have a direct counterpart in the |VectorArray| interface.
    """

    @abstractmethod
    def copy(self, deep=False):
        pass

    @abstractmethod
    def scal(self, alpha):
        pass

    @abstractmethod
    def axpy(self, alpha, x):
        pass

    def inner(self, other):
        raise NotImplementedError

    @abstractmethod
    def norm(self):
        pass

    @abstractmethod
    def norm2(self):
        pass

    def sup_norm(self):
        _, max_val = self.amax()
        return max_val

    @abstractmethod
    def dofs(self, dof_indices):
        pass

    @abstractmethod
    def amax(self):
        pass

    def __add__(self, other):
        result = self.copy()
        result.axpy(1, other)
        return result

    def __iadd__(self, other):
        self.axpy(1, other)
        return self

    __radd__ = __add__

    def __sub__(self, other):
        result = self.copy()
        result.axpy(-1, other)
        return result

    def __isub__(self, other):
        self.axpy(-1, other)
        return self

    def __mul__(self, other):
        result = self.copy()
        result.scal(other)
        return result

    def __imul__(self, other):
        self.scal(other)
        return self

    def __neg__(self):
        result = self.copy()
        result.scal(-1)
        return result

    @property
    def real(self):
        return self.copy()

    @property
    def imag(self):
        return None

    def conj(self):
        return self.copy()


class CopyOnWriteVector(Vector):

    @abstractclassmethod
    def from_instance(cls, instance):
        pass

    @abstractmethod
    def _copy_data(self):
        pass

    @abstractmethod
    def _scal(self, alpha):
        pass

    @abstractmethod
    def _axpy(self, alpha, x):
        pass

    def copy(self, deep=False):
        c = self.from_instance(self)
        if deep:
            c._copy_data()
        else:
            try:
                self._refcount[0] += 1
            except AttributeError:
                self._refcount = [2]
            c._refcount = self._refcount
        return c

    def scal(self, alpha):
        self._copy_data_if_needed()
        self._scal(alpha)

    def axpy(self, alpha, x):
        self._copy_data_if_needed()
        self._axpy(alpha, x)

    def __del__(self):
        try:
            self._refcount[0] -= 1
        except AttributeError:
            pass

    def _copy_data_if_needed(self):
        try:
            if self._refcount[0] > 1:
                self._refcount[0] -= 1
                self._copy_data()
                self._refcount = [1]
        except AttributeError:
            self._refcount = [1]


class ComplexifiedVector(Vector):

    def __init__(self, real_part, imag_part):
        self.real_part, self.imag_part = real_part, imag_part

    def copy(self, deep=False):
        real_part = self.real_part.copy(deep=deep)
        imag_part = self.imag_part.copy(deep=deep) if self.imag_part is not None else None
        return type(self)(real_part, imag_part)

    def scal(self, alpha):
        if self.imag_part is None:
            self.real_part.scal(alpha.real)
            if alpha.imag != 0:
                self.imag_part = self.real_part * alpha.imag
        else:
            if alpha.imag == 0:
                self.real_part.scal(alpha.real)
                self.imag_part.scal(alpha.real)
            else:
                old_real_part = self.real_part.copy()
                self.real_part.scal(alpha.real)
                self.real_part.axpy(-alpha.imag, self.imag_part)
                self.imag_part.scal(alpha.real)
                self.imag_part.axpy(alpha.imag, old_real_part)

    def axpy(self, alpha, x):
        if x is self:
            self.scal(1. + alpha)
            return

        # real part
        self.real_part.axpy(alpha.real, x.real_part)
        if x.imag_part is not None:
            self.real_part.axpy(-alpha.imag, x.imag_part)

        # imaginary part
        if alpha.imag != 0:
            if self.imag_part is None:
                self.imag_part = x.real_part * alpha.imag
            else:
                self.imag_part.axpy(alpha.imag, x.real_part)
        if x.imag_part is not None:
            if self.imag_part is None:
                self.imag_part = x.imag_part * alpha.real
            else:
                self.imag_part.axpy(alpha.real, x.imag_part)

    def inner(self, other):
        result = self.real_part.inner(other.real_part)
        if self.imag_part is not None:
            result += self.imag_part.inner(other.real_part) * (-1j)
        if other.imag_part is not None:
            result += self.real_part.inner(other.imag_part) * 1j
        if self.imag_part is not None and other.imag_part is not None:
            result += self.imag_part.inner(other.imag_part)
        return result

    def norm(self):
        result = self.real_part.norm()
        if self.imag_part is not None:
            result = np.linalg.norm([result, self.imag_part.norm()])
        return result

    def norm2(self):
        result = self.real_part.norm2()
        if self.imag_part is not None:
            result += self.imag_part.norm2()
        return result

    def sup_norm(self):
        if self.imag_part is not None:
            # we cannot compute the sup_norm from the sup_norms of real_part and imag_part
            return self.amax()[1]
        return self.real_part.sup_norm()

    def dofs(self, dof_indices):
        values = self.real_part.dofs(dof_indices)
        if self.imag_part is not None:
            imag_values = self.imag_part.dofs(dof_indices)
            return values + imag_values * 1j
        else:
            return values

    def amax(self):
        if self.imag_part is not None:
            raise NotImplementedError
        return self.real_part.amax()

    def to_numpy(self, ensure_copy=False):
        if self.imag_part is not None:
            return self.real_part.to_numpy(ensure_copy=False) + self.imag_part.to_numpy(ensure_copy=False) * 1j
        else:
            return self.real_part.to_numpy(ensure_copy=ensure_copy)

    @property
    def real(self):
        return type(self)(self.real_part.copy(), None)

    @property
    def imag(self):
        return type(self)(self.imag_part.copy(), None) if self.imag_part is not None else None

    def conj(self):
        return type(self)(self.real_part.copy(), -self.imag_part if self.imag_part is not None else None)


class NumpyVector(CopyOnWriteVector):
    """Vector stored in a NumPy 1D-array."""

    def __init__(self, array):
        self._array = array

    @classmethod
    def from_instance(cls, instance):
        return cls(instance._array)

    def to_numpy(self, ensure_copy=False):
        if ensure_copy:
            return self._array.copy()
        else:
            self._copy_data_if_needed()
            return self._array

    @property
    def dim(self):
        return len(self._array)

    def _copy_data(self):
        self._array = self._array.copy()

    def _scal(self, alpha):
        try:
            self._array *= alpha
        except TypeError:  # e.g. when scaling real array by complex alpha
            self._array = self._array * alpha

    def _axpy(self, alpha, x):
        assert self.dim == x.dim
        if alpha == 0:
            return
        if alpha == 1:
            try:
                self._array += x._array
            except TypeError:
                self._array = self._array + x._array
        elif alpha == -1:
            try:
                self._array -= x._array
            except TypeError:
                self._array = self._array - x._array
        else:
            try:
                self._array += x._array * alpha
            except TypeError:
                self._array = self._array + x._array * alpha

    def inner(self, other):
        assert self.dim == other.dim
        return np.sum(self._array.conj() * other._array)

    def norm(self):
        return np.linalg.norm(self._array)

    def norm2(self):
        return np.sum((self._array * self._array.conj()).real)

    def dofs(self, dof_indices):
        return self._array[dof_indices]

    def amax(self):
        A = np.abs(self._array)
        max_ind = np.argmax(A)
        max_val = A[max_ind]
        return max_ind, np.abs(max_val)

    @property
    def real(self):
        return self.__class__(self._array.real.copy())

    @property
    def imag(self):
        return self.__class__(self._array.imag.copy())

    def conj(self):
        return self.__class__(self._array.conj())


class ListVectorArray(VectorArray):
    """|VectorArray| implemented as a Python list of vectors.

    This |VectorArray| implementation is the first choice when
    creating pyMOR wrappers for external solvers which are based
    on single vector objects. In order to do so, a wrapping
    subclass of :class:`Vector` has to be provided
    on which the implementation of |ListVectorArray| will operate.
    The associated |VectorSpace| is a subclass of
    :class:`ListVectorSpace`.

    For an example, see :class:`NumpyVector`, :class:`NumpyListVectorSpace`
    or :class:`~pymor.bindings.fenics.FenicsVector`,
    :class:`~pymor.bindings.fenics.FenicsVectorSpace`.
    """

    _NONE = ()

    def __init__(self, vectors, space):
        self._list = vectors
        self.space = space

    def to_numpy(self, ensure_copy=False):
        if len(self._list) > 0:
            return np.array([v.to_numpy() for v in self._list])
        else:
            return np.empty((0, self.dim))

    @property
    def _data(self):
        """Return list of NumPy Array views on vector data for hacking / interactive use."""
        return ListVectorArrayNumpyView(self)

    def __len__(self):
        return len(self._list)

    def __getitem__(self, ind):
        if isinstance(ind, Number) and (ind >= len(self) or ind < -len(self)):
            raise IndexError('VectorArray index out of range')
        return ListVectorArrayView(self, ind)

    def __delitem__(self, ind):
        assert self.check_ind(ind)
        if hasattr(ind, '__len__'):
            thelist = self._list
            l = len(thelist)
            remaining = sorted(set(range(l)) - {i if 0 <= i else l+i for i in ind})
            self._list = [thelist[i] for i in remaining]
        else:
            del self._list[ind]

    def append(self, other, remove_from_other=False):
        assert other.space == self.space
        assert not remove_from_other or (other is not self and getattr(other, 'base', None) is not self)

        if not remove_from_other:
            self._list.extend([v.copy() for v in other._list])
        else:
            self._list.extend(other._list)
            if other.is_view:
                del other.base[other.ind]
            else:
                del other[:]

    def copy(self, deep=False):
        return ListVectorArray([v.copy(deep=deep) for v in self._list], self.space)

    def scal(self, alpha):
        assert isinstance(alpha, Number) \
            or isinstance(alpha, np.ndarray) and alpha.shape == (len(self),)

        if type(alpha) is np.ndarray:
            for a, v in zip(alpha, self._list):
                v.scal(a)
        else:
            for v in self._list:
                v.scal(alpha)

    def axpy(self, alpha, x):
        assert self.space == x.space
        len_x = len(x)
        assert len(self) == len_x or len_x == 1
        assert isinstance(alpha, Number) \
            or isinstance(alpha, np.ndarray) and alpha.shape == (len(self),)

        if np.all(alpha == 0):
            return

        if self is x or x.is_view and self is x.base:
            x = x.copy()

        if len(x) == 1:
            xx = x._list[0]
            if type(alpha) is np.ndarray:
                for a, y in zip(alpha, self._list):
                    y.axpy(a, xx)
            else:
                for y in self._list:
                    y.axpy(alpha, xx)
        else:
            if type(alpha) is np.ndarray:
                for a, xx, y in zip(alpha, x._list, self._list):
                    y.axpy(a, xx)
            else:
                for xx, y in zip(x._list, self._list):
                    y.axpy(alpha, xx)

    def inner(self, other, product=None):
        assert self.space == other.space
        if product is not None:
            return product.apply2(self, other)

        return np.array([[a.inner(b) for b in other._list] for a in self._list]).reshape((len(self), len(other)))

    def pairwise_inner(self, other, product=None):
        assert self.space == other.space
        assert len(self._list) == len(other)
        if product is not None:
            return product.pairwise_apply2(self, other)

        return np.array([a.inner(b) for a, b in zip(self._list, other._list)])

    def gramian(self, product=None):
        if product is not None:
            return super().gramian(product)
        l = len(self._list)
        R = [[0.] * l for _ in range(l)]
        for i in range(l):
            for j in range(i, l):
                R[i][j] = self._list[i].inner(self._list[j])
                if i == j:
                    R[i][j] = R[i][j].real
                else:
                    R[j][i] = R[i][j].conjugate()
        R = np.array(R)
        return R

    def lincomb(self, coefficients):
        assert 1 <= coefficients.ndim <= 2
        if coefficients.ndim == 1:
            coefficients = coefficients[np.newaxis, :]

        assert coefficients.shape[1] == len(self)

        RL = []
        for coeffs in coefficients:
            R = self.space.zero_vector()
            for v, c in zip(self._list, coeffs):
                R.axpy(c, v)
            RL.append(R)

        return ListVectorArray(RL, self.space)

    def _norm(self):
        return np.array([v.norm() for v in self._list])

    def _norm2(self):
        return np.array([v.norm2() for v in self._list])

    def sup_norm(self):
        if self.dim == 0:
            return np.zeros(len(self))
        else:
            return np.array([v.sup_norm() for v in self._list])

    def dofs(self, dof_indices):
        assert isinstance(dof_indices, list) and (len(dof_indices) == 0 or min(dof_indices) >= 0) \
            or (isinstance(dof_indices, np.ndarray) and dof_indices.ndim == 1
                and (len(dof_indices) == 0 or np.min(dof_indices) >= 0))
        assert len(self) > 0 or len(dof_indices) == 0 or max(dof_indices) < self.dim
        return np.array([v.dofs(dof_indices) for v in self._list]).reshape((len(self), len(dof_indices)))

    def amax(self):
        assert self.dim > 0

        MI = np.empty(len(self._list), dtype=int)
        MV = np.empty(len(self._list))

        for k, v in enumerate(self._list):
            MI[k], MV[k] = v.amax()

        return MI, MV

    @property
    def real(self):
        return ListVectorArray([v.real for v in self._list], self.space)

    @property
    def imag(self):
        # note that Vector.imag is allowed to return None in case
        # of a real vector, so we have to check for that.
        # returning None is allowed as ComplexifiedVector does not know
        # how to create a new zero vector.
        return ListVectorArray([v.imag or self.space.zero_vector() for v in self._list], self.space)

    def conj(self):
        return self.__class__([v.conj() for v in self._list], self.space)

    def __str__(self):
        return f'{type(self).__name__} of {len(self._list)} vectors of space {self.space}'


class ListVectorSpace(VectorSpace):
    """|VectorSpace| of |ListVectorArrays|."""

    dim = None
    vector_type = Vector

    @abstractmethod
    def zero_vector(self):
        pass

    def ones_vector(self):
        return self.full_vector(1.)

    def full_vector(self, value):
        return self.vector_from_numpy(np.full(self.dim, value))

    def random_vector(self, distribution, random_state, **kwargs):
        values = _create_random_values(self.dim, distribution, random_state, **kwargs)
        return self.vector_from_numpy(values)

    @abstractmethod
    def make_vector(self, obj):
        pass

    def vector_from_numpy(self, data, ensure_copy=False):
        raise NotImplementedError

    @classmethod
    def space_from_vector_obj(cls, vec, id):
        raise NotImplementedError

    @classmethod
    def space_from_dim(cls, dim, id):
        raise NotImplementedError

    def zeros(self, count=1, reserve=0):
        assert count >= 0 and reserve >= 0
        return ListVectorArray([self.zero_vector() for _ in range(count)], self)

    def ones(self, count=1, reserve=0):
        assert count >= 0 and reserve >= 0
        return ListVectorArray([self.ones_vector() for _ in range(count)], self)

    def full(self, value, count=1, reserve=0):
        assert count >= 0 and reserve >= 0
        return ListVectorArray([self.full_vector(value) for _ in range(count)], self)

    def random(self, count=1, distribution='uniform', random_state=None, seed=None, reserve=0, **kwargs):
        assert count >= 0 and reserve >= 0
        assert random_state is None or seed is None
        random_state = get_random_state(random_state, seed)
        return ListVectorArray([self.random_vector(distribution=distribution, random_state=random_state, **kwargs)
                                for _ in range(count)], self)

    @classinstancemethod
    def make_array(cls, obj, id=None):
        if len(obj) == 0:
            raise NotImplementedError
        return cls.space_from_vector_obj(obj[0], id=id).make_array(obj)

    @make_array.instancemethod
    def make_array(self, obj):
        """:noindex:"""
        return ListVectorArray([v if isinstance(v, self.vector_type) else self.make_vector(v) for v in obj], self)

    @classinstancemethod
    def from_numpy(cls, data, id=None, ensure_copy=False):
        return cls.space_from_dim(data.shape[1], id=id).from_numpy(data, ensure_copy=ensure_copy)

    @from_numpy.instancemethod
    def from_numpy(self, data, ensure_copy=False):
        """:noindex:"""
        return ListVectorArray([self.vector_from_numpy(v, ensure_copy=ensure_copy) for v in data], self)


class ComplexifiedListVectorSpace(ListVectorSpace):

    real_vector_type = Vector
    vector_type = ComplexifiedVector

    @abstractmethod
    def real_zero_vector(self):
        pass

    def zero_vector(self):
        return self.vector_type(self.real_zero_vector(), None)

    def real_full_vector(self, value):
        return self.real_vector_from_numpy(np.full(self.dim, value))

    def full_vector(self, value):
        return self.vector_type(self.real_full_vector(value), None)

    def real_random_vector(self, distribution, random_state, **kwargs):
        values = _create_random_values(self.dim, distribution, random_state, **kwargs)
        return self.real_vector_from_numpy(values)

    def random_vector(self, distribution, random_state, **kwargs):
        return self.vector_type(self.real_random_vector(distribution, random_state, **kwargs), None)

    @abstractmethod
    def real_make_vector(self, obj):
        pass

    def make_vector(self, obj):
        if isinstance(obj, self.real_vector_type):
            return self.vector_type(obj, None)
        else:
            return self.vector_type(self.real_make_vector(obj), None)

    def real_vector_from_numpy(self, data, ensure_copy=False):
        raise NotImplementedError

    def vector_from_numpy(self, data, ensure_copy=False):
        if np.iscomplexobj(data):
            real_part = self.real_vector_from_numpy(data.real)
            imag_part = self.real_vector_from_numpy(data.imag)
        else:
            real_part = self.real_vector_from_numpy(data, ensure_copy=ensure_copy)
            imag_part = None
        return self.vector_type(real_part, imag_part)


class NumpyListVectorSpace(ListVectorSpace):

    vector_type = NumpyVector

    def __init__(self, dim, id=None):
        self.dim = dim
        self.id = id

    def __eq__(self, other):
        return type(other) is NumpyListVectorSpace and self.dim == other.dim and self.id == other.id

    @classmethod
    def space_from_vector_obj(cls, vec, id):
        return cls(len(vec), id)

    @classmethod
    def space_from_dim(cls, dim, id):
        return cls(dim, id)

    def zero_vector(self):
        return NumpyVector(np.zeros(self.dim))

    def ones_vector(self):
        return NumpyVector(np.ones(self.dim))

    def full_vector(self, value):
        return NumpyVector(np.full(self.dim, value))

    def make_vector(self, obj):
        obj = np.asarray(obj)
        assert obj.ndim == 1 and len(obj) == self.dim
        return NumpyVector(obj)

    def vector_from_numpy(self, data, ensure_copy=False):
        return self.make_vector(data.copy() if ensure_copy else data)


class ListVectorArrayView(ListVectorArray):

    is_view = True

    def __init__(self, base, ind):
        self.base = base
        assert base.check_ind(ind)
        self.ind = base.normalize_ind(ind)
        if type(ind) is slice:
            self._list = base._list[ind]
        elif hasattr(ind, '__len__'):
            _list = base._list
            self._list = [_list[i] for i in ind]
        else:
            self._list = [base._list[ind]]

    @property
    def space(self):
        return self.base.space

    def __getitem__(self, ind):
        try:
            return self.base[self.base.sub_index(self.ind, ind)]
        except IndexError:
            raise IndexError('VectorArray index out of range')

    def __delitem__(self, ind):
        raise TypeError('Cannot remove from ListVectorArrayView')

    def append(self, other, remove_from_other=False):
        raise TypeError('Cannot append to ListVectorArrayView')

    def scal(self, alpha):
        assert self.base.check_ind_unique(self.ind)
        super().scal(alpha)

    def axpy(self, alpha, x):
        assert self.base.check_ind_unique(self.ind)
        if x is self.base or x.is_view and x.base is self.base:
            x = x.copy()
        super().axpy(alpha, x)


class ListVectorArrayNumpyView:

    def __init__(self, array):
        self.array = array

    def __len__(self):
        return len(self.array)

    def __getitem__(self, i):
        return self.array._list[i].to_numpy()

    def __repr__(self):
        return '[' + ',\n '.join(repr(v) for v in self) + ']'



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import numpy as np

from pymortests.base import runmodule
from pymor.discretizers.builtin import RectGrid, TriaGrid


def test_with_newtype():
    g = RectGrid(num_intervals=(99, 99))
    g2 = g.with_(new_type=TriaGrid, domain=([0, 0], [2, 2]))

    assert isinstance(g2, TriaGrid)
    assert g2.num_intervals == (99, 99)
    assert np.all(g2.domain == ([0, 0], [2, 2]))


if __name__ == "__main__":
    runmodule(filename=__file__)



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import asyncio
from collections import defaultdict
from math import ceil
import numpy as np
from threading import Thread

from pymor.algorithms.pod import pod
from pymor.core.logger import getLogger


class Node:

    def __init__(self, tag=None, parent=None, after=None):
        after = after or []
        self.tag, self.parent, self.after = tag, parent, after
        self.children = []
        if parent:
            parent.children.append(self)

    def add_child(self, tag=None, after=None, **kwargs):
        return Node(tag=tag, parent=self, after=after, **kwargs)

    @property
    def depth(self):
        max_level = 0
        for _, level in self.traverse(True):
            max_level = max(max_level, level)
        return max_level + 1

    @property
    def is_leaf(self):
        return not self.children

    @property
    def is_root(self):
        return not self.parent

    def traverse(self, return_level=False):
        current_node = self
        last_child = None
        level = 0
        while True:
            if last_child is None:
                if return_level:
                    yield current_node, level
                else:
                    yield current_node
            if current_node.children:
                if last_child is None:
                    current_node = current_node.children[0]
                    level += 1
                    continue
                else:
                    last_child_pos = current_node.children.index(last_child)
                    if last_child_pos + 1 < len(current_node.children):
                        current_node = current_node.children[last_child_pos+1]
                        last_child = None
                        level += 1
                        continue
            if not current_node.parent:
                return
            last_child = current_node
            current_node = current_node.parent
            level -= 1

    def __str__(self):
        lines = []
        for node, level in self.traverse(True):
            line = ''
            if node.parent:
                p = node.parent
                while p.parent:
                    if p.parent.children.index(p) + 1 == len(p.parent.children):
                        line = '  ' + line
                    else:
                        line = '| ' + line
                    p = p.parent
                line += '+-o'
            else:
                line += 'o'
            if node.tag is not None:
                line += f' {node.tag}'
            if node.after:
                line += f' (after {",".join(str(a) for a in node.after)})'
            lines.append(line)
        return '\n'.join(lines)


def inc_hapod_tree(steps):
    tree = node = Node()
    for step in range(steps)[::-1]:
        # add leaf node for a new snapshot
        node.add_child(tag=step, after=(step-1,) if step > 0 else None)
        if step > 0:
            # add node for the previous POD step
            node = node.add_child()
    return tree


def dist_hapod_tree(num_slices, arity=None):
    tree = Node()
    if arity is None:
        arity = num_slices

    def add_children(node, slices):
        if len(slices) > arity:
            sub_slices = np.array_split(slices, arity)
            for s in sub_slices:
                if len(s) > 1:
                    child = node.add_child()
                    add_children(child, s)
                else:
                    child = node.add_child(s.item())
        else:
            for s in slices:
                node.add_child(s)

    add_children(tree, np.arange(num_slices))

    return tree


def default_pod_method(U, eps, is_root_node, product):
    return pod(U, atol=0., rtol=0.,
               l2_err=eps, product=product,
               orth_tol=None if is_root_node else np.inf)


def hapod(tree, snapshots, local_eps, product=None, pod_method=default_pod_method,
          executor=None, eval_snapshots_in_executor=False):
    """Compute the Hierarchical Approximate POD.

    This is an implementation of the HAPOD algorithm from :cite:`HLR18`.

    Parameters
    ----------
    tree
        A :class:`Tree` defining the worker topology.
    snapshots
        A mapping `snapshots(node)` returning for each leaf node the
        associated snapshot vectors.
    local_eps
        A mapping `local_eps(node, snap_count, num_vecs)` assigning
        to each tree node `node` an l2 truncation error tolerance for the
        local pod based on the number of input vectors `num_vecs` and the
        total number of snapshot vectors below the given node `snap_count`.
    product
        Inner product |Operator| w.r.t. which to compute the POD.
    pod_method
        A function `pod_method(U, eps, root_node, product)` for computing
        the POD of the |VectorArray| `U` w.r.t. the given inner product
        `product` and the l2 error tolerance `eps`. `root_node` is set to
        `True` when the POD is computed at the root of the tree.
    executor
        If not `None`, a :class:`concurrent.futures.Executor` object to use
        for parallelization.
    eval_snapshots_in_executor
        If `True` also parallelize the evaluation of the snapshot map.

    Returns
    -------
    modes
        The computed POD modes.
    svals
        The associated singular values.
    snap_count
        The total number of input snapshot vectors.
    """
    logger = getLogger('pymor.algorithms.hapod.hapod')

    node_finished_events = defaultdict(asyncio.Event)

    async def hapod_step(node):
        if node.after:
            await asyncio.wait([node_finished_events[a].wait() for a in node.after])

        if node.children:
            modes, svals, snap_counts = zip(
                *await asyncio.gather(*(hapod_step(c) for c in node.children))
            )
            for m, sv in zip(modes, svals):
                m.scal(sv)
            U = modes[0]
            for V in modes[1:]:
                U.append(V, remove_from_other=True)
            snap_count = sum(snap_counts)
        else:
            logger.info(f'Obtaining snapshots for node {node.tag or ""} ...')
            if eval_snapshots_in_executor:
                U = await executor.submit(snapshots, node)
            else:
                U = snapshots(node)
            snap_count = len(U)

        eps = local_eps(node, snap_count, len(U))
        if eps:
            logger.info('Computing intermediate POD ...')
            modes, svals = await executor.submit(pod_method, U, eps, not node.parent, product)
        else:
            modes, svals = U.copy(), np.ones(len(U))
        if node.tag is not None:
            node_finished_events[node.tag].set()
        return modes, svals, snap_count

    # wrap Executer to ensure LIFO ordering of tasks
    # this ensures that PODs of parent nodes are computed as soon as all input data
    # is available
    if executor is not None:
        executor = LifoExecutor(executor)
    else:
        executor = FakeExecutor

    # run new asyncio event loop in separate thread to not interfere with
    # already running event loops (e.g. jupyter)

    def main():
        nonlocal result
        result = asyncio.run(hapod_step(tree))
    result = None
    hapod_thread = Thread(target=main)
    hapod_thread.start()
    hapod_thread.join()
    return result


def inc_hapod(steps, snapshots, eps, omega, product=None, executor=None):
    """Incremental Hierarchical Approximate POD.

    This computes the incremental HAPOD from :cite:`HLR18`.

    Parameters
    ----------
    steps
        The number of incremental POD updates. Has to agree with the length
        of `snapshots`.
    snapshots
        An iterable returning for each incremental POD step the associated
        snapshot vectors.
    eps
        Desired l2-mean approximation error.
    omega
        Tuning parameter (0 < omega < 1) to balance performance with
        approximation quality.
    product
        Inner product |Operator| w.r.t. which to compute the POD.
    executor
        If not `None`, a :class:`concurrent.futures.Executor` object to use
        to compute new snapshot vectors and POD updates in parallel.

    Returns
    -------
    modes
        The computed POD modes.
    svals
        The associated singular values.
    snap_count
        The total number of input snapshot vectors.
    """
    tree = inc_hapod_tree(steps)

    last_step = -1
    snapshots = iter(snapshots)

    def get_snapshots(node):
        nonlocal last_step
        assert node.tag == last_step + 1
        last_step += 1
        return next(snapshots)

    result = hapod(tree,
                   get_snapshots,
                   std_local_eps(tree, eps, omega, False),
                   product=product,
                   executor=executor,
                   eval_snapshots_in_executor=True)
    assert last_step == steps - 1
    return result


def dist_hapod(num_slices, snapshots, eps, omega, arity=None,
               product=None, executor=None, eval_snapshots_in_executor=False):
    """Distributed Hierarchical Approximate POD.

    This computes the distributed HAPOD from :cite:`HLR18`.

    Parameters
    ----------
    num_slices
        The number of snapshot vector slices.
    snapshots
        A mapping `snapshots(slice)` returning for each slice number
        the associated snapshot vectors.
    eps
        Desired l2-mean approximation error.
    omega
        Tuning parameter (0 < omega < 1) to balance performance with
        approximation quality.
    arity
        If not `None`, the arity of the HAPOD tree. Otherwise, a
        tree of depth 2 is used (one POD per slice and one additional
        POD of the resulting data).
    product
        Inner product |Operator| w.r.t. which to compute the POD.
    executor
        If not `None`, a :class:`concurrent.futures.Executor` object to use
        for parallelization.
    eval_snapshots_in_executor
        If `True` also parallelize the evaluation of the snapshot map.

    Returns
    -------
    modes
        The computed POD modes.
    svals
        The associated singular values.
    snap_count
        The total number of input snapshot vectors.
    """
    tree = dist_hapod_tree(num_slices, arity)
    return hapod(tree,
                 snapshots,
                 std_local_eps(tree, eps, omega, True),
                 product=product, executor=executor,
                 eval_snapshots_in_executor=eval_snapshots_in_executor)


def inc_vectorarray_hapod(steps, U, eps, omega, product=None):
    """Incremental Hierarchical Approximate POD.

    This computes the incremental HAPOD from :cite:`HLR18` for a given |VectorArray|.

    Parameters
    ----------
    steps
        The number of incremental POD updates.
    U
        The |VectorArray| of which to compute the HAPOD.
    eps
        Desired l2-mean approximation error.
    omega
        Tuning parameter (0 < omega < 1) to balance performance with
        approximation quality.
    product
        Inner product |Operator| w.r.t. which to compute the POD.

    Returns
    -------
    modes
        The computed POD modes.
    svals
        The associated singular values.
    snap_count
        The total number of input snapshot vectors.
    """
    chunk_size = ceil(len(U) / steps)
    slices = range(0, len(U), chunk_size)

    def snapshots():
        for slice in slices:
            yield U[slice: slice+chunk_size]

    return inc_hapod(len(slices), snapshots(),
                     eps, omega, product=product)


def dist_vectorarray_hapod(num_slices, U, eps, omega, arity=None, product=None, executor=None):
    """Distributed Hierarchical Approximate POD.

    This computes the distributed HAPOD from :cite:`HLR18` of a given |VectorArray|.

    Parameters
    ----------
    num_slices
        The number of snapshot vector slices.
    U
        The |VectorArray| of which to compute the HAPOD.
    eps
        Desired l2-mean approximation error.
    omega
        Tuning parameter (0 < omega < 1) to balance performance with
        approximation quality.
    arity
        If not `None`, the arity of the HAPOD tree. Otherwise, a
        tree of depth 2 is used (one POD per slice and one additional
        POD of the resulting data).
    product
        Inner product |Operator| w.r.t. which to compute the POD.
    executor
        If not `None`, a :class:`concurrent.futures.Executor` object to use
        for parallelization.

    Returns
    -------
    modes
        The computed POD modes.
    svals
        The associated singular values.
    snap_count
        The total number of input snapshot vectors.
    """
    chunk_size = ceil(len(U) / num_slices)
    slices = range(0, len(U), chunk_size)
    return dist_hapod(len(slices),
                      lambda i: U[slices[i.tag]: slices[i.tag]+chunk_size],
                      eps, omega, arity=arity, product=product, executor=executor)


def std_local_eps(tree, eps, omega, pod_on_leafs=True):

    L = tree.depth if pod_on_leafs else tree.depth - 1

    def local_eps(node, snap_count, input_count):
        if node.is_root:
            return np.sqrt(snap_count) * omega * eps
        elif node.is_leaf and not pod_on_leafs:
            return 0.
        else:
            return np.sqrt(snap_count) / np.sqrt(L - 1) * np.sqrt(1 - omega**2) * eps

    return local_eps


class LifoExecutor:

    def __init__(self, executor, max_workers=None):
        self.executor = executor
        self.max_workers = max_workers or executor._max_workers
        self.queue = []

    def submit(self, f, *args):
        future = asyncio.get_event_loop().create_future()
        self.queue.append((future, f, args))
        asyncio.get_event_loop().create_task(self.run_task())
        return future

    async def run_task(self):
        if not hasattr(self, 'sem'):
            self.sem = asyncio.Semaphore(self.max_workers)
        await self.sem.acquire()
        future, f, args = self.queue.pop()
        executor_future = asyncio.get_event_loop().run_in_executor(self.executor, f, *args)
        executor_future.add_done_callback(lambda f, ff=future: self.done_callback(future, f))

    def done_callback(self, future, executor_future):
        self.sem.release()
        future.set_result(executor_future.result())


class FakeExecutor:

    @staticmethod
    async def submit(f, *args):
        return f(*args)



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

from numbers import Number

import numpy as np
from scipy.linalg import solve, solve_triangular

from pymor.analyticalproblems.expressions import parse_expression
from pymor.core.base import abstractmethod
from pymor.parameters.base import ParametricObject, Mu
from pymor.parameters.functionals import ParameterFunctional


class Function(ParametricObject):
    """Interface for |Parameter| dependent analytical functions.

    Every |Function| is a map of the form ::

       f(μ): Ω ⊆ R^d -> R^(shape_range)

    The returned values are |NumPy arrays| of arbitrary (but fixed)
    shape. Note that NumPy distinguishes between one-dimensional
    arrays of length 1 (with shape `(1,)`) and zero-dimensional
    scalar arrays (with shape `()`). In pyMOR, we usually
    expect scalar-valued functions to have `shape_range == ()`.

    While the function might raise an error if it is evaluated
    for an argument not in the domain Ω, the exact behavior is left
    undefined.

    Functions are vectorized in the sense, that if `x.ndim == k`, then ::

       f(x, μ)[i0, i1, ..., i(k-2)] == f(x[i0, i1, ..., i(k-2)], μ).

    In particular, `f(x, μ).shape == x.shape[:-1] + shape_range`.

    Attributes
    ----------
    dim_domain
        The dimension d > 0.
    shape_range
        The shape of the function values.
    """

    @abstractmethod
    def evaluate(self, x, mu=None):
        """Evaluate the function for given argument `x` and |parameter values| `mu`."""
        pass

    def __call__(self, x, mu=None):
        """Shorthand for :meth:`~Function.evaluate`."""
        return self.evaluate(x, mu)

    def _add_sub(self, other, sign):
        if isinstance(other, Number) and other == 0:
            return self
        elif not isinstance(other, Function):
            other = np.array(other)
            assert other.shape == self.shape_range
            if np.all(other == 0.):
                return self
            other = ConstantFunction(other, dim_domain=self.dim_domain)

        if self.name != 'LincombFunction' or not isinstance(self, LincombFunction):
            if other.name == 'LincombFunction' and isinstance(other, LincombFunction):
                functions = (self,) + other.functions
                coefficients = (1.,) + (other.coefficients if sign == 1. else tuple(-c for c in other.coefficients))
            else:
                functions, coefficients = (self, other), (1., sign)
        elif other.name == 'LincombFunction' and isinstance(other, LincombFunction):
            functions = self.functions + other.functions
            coefficients = self.coefficients + (other.coefficients if sign == 1.
                                                else tuple(-c for c in other.coefficients))
        else:
            functions, coefficients = self.functions + (other,), self.coefficients + (sign,)

        return LincombFunction(functions, coefficients)

    def _radd_sub(self, other, sign):
        assert not isinstance(other, Function)  # handled by __add__/__sub__
        if isinstance(other, Number) and other == 0:
            return self

        other = np.array(other)
        assert other.shape == self.shape_range
        if np.all(other == 0.):
            return self
        other = ConstantFunction(other, dim_domain=self.dim_domain)

        if self.name != 'LincombFunction' or not isinstance(self, LincombFunction):
            functions, coefficients = (other, self), (1., sign)
        else:
            functions = (other,) + self.functions
            coefficients = (1.,) + (self.coefficients if sign == 1. else tuple(-c for c in self.coefficients))

        return LincombFunction(functions, coefficients)

    def __add__(self, other):
        return self._add_sub(other, 1.)

    def __sub__(self, other):
        return self._add_sub(other, -1.)

    def __radd__(self, other):
        return self._radd_sub(other, 1.)

    def __rsub__(self, other):
        return self._radd_sub(other, -1.)

    def __mul__(self, other):
        if not isinstance(other, (Number, ParameterFunctional, Function)):
            return NotImplemented
        if isinstance(other, (Number, ParameterFunctional)):
            return LincombFunction([self], [other])
        if self.name != 'ProductFunction' or not isinstance(self, ProductFunction):
            if isinstance(other, ProductFunction) and other.name == 'ProductFunction':
                return other.with_(functions=other.functions + [self])
            else:
                return ProductFunction([self, other])
        elif isinstance(other, ProductFunction) and other.name == 'ProductFunction':
            functions = self.functions + other.functions
            return ProductFunction(functions)
        else:
            return self.with_(functions=self.functions + [other])

    __rmul__ = __mul__

    def __neg__(self):
        return LincombFunction([self], [-1.])


class ConstantFunction(Function):
    """A constant |Function|

    Defined as ::

        f: R^d -> R^shape(c), f(x) = c

    Parameters
    ----------
    value
        The constant c.
    dim_domain
        The dimension d.
    name
        The name of the function.
    """

    def __init__(self, value=np.array(1.0), dim_domain=1, name=None):
        assert dim_domain > 0
        assert isinstance(value, (Number, np.ndarray))
        value = np.array(value)
        self.__auto_init(locals())
        self.shape_range = value.shape

    def __str__(self):
        return f'{self.name}: x -> {self.value}'

    def evaluate(self, x, mu=None):
        x = np.array(x, copy=False, ndmin=1)
        assert x.shape[-1] == self.dim_domain
        if x.ndim == 1:
            return np.array(self.value)
        else:
            return np.tile(self.value, x.shape[:-1] + (1,) * len(self.shape_range))


class GenericFunction(Function):
    """Wrapper making an arbitrary Python function between |NumPy arrays| a proper |Function|.

    Note that a :class:`GenericFunction` can only be :mod:`pickled <pymor.core.pickle>`
    if the function it is wrapping can be pickled (cf. :func:`~pymor.core.pickle.dumps_function`).
    For this reason, it is usually preferable to use :class:`ExpressionFunction`
    instead of :class:`GenericFunction`.

    Parameters
    ----------
    mapping
        The function to wrap. If `parameters` is `None`, the function is of
        the form `mapping(x)`. If `parameters` is not `None`, the function has
        to have the signature `mapping(x, mu)`. Moreover, the function is expected
        to be vectorized, i.e.::

            mapping(x).shape == x.shape[:-1] + shape_range.

    dim_domain
        The dimension of the domain.
    shape_range
        The shape of the values returned by the mapping.
    parameters
        The |Parameters| the mapping accepts.
    name
        The name of the function.
    """

    def __init__(self, mapping, dim_domain=1, shape_range=(), parameters={}, name=None):
        assert dim_domain > 0
        assert isinstance(shape_range, (Number, tuple))
        if not isinstance(shape_range, tuple):
            shape_range = (shape_range,)
        self.parameters_own = parameters
        self.__auto_init(locals())

    def __str__(self):
        return f'{self.name}: x -> {self.mapping}'

    def evaluate(self, x, mu=None):
        assert self.parameters.assert_compatible(mu)
        x = np.array(x, copy=False, ndmin=1)
        assert x.shape[-1] == self.dim_domain

        if self.parametric:
            v = self.mapping(x, mu=mu)
        else:
            v = self.mapping(x)

        if v.shape != x.shape[:-1] + self.shape_range:
            assert v.shape[:len(x.shape) - 1] == x.shape[:-1]
            v = v.reshape(x.shape[:-1] + self.shape_range)

        return v


class ExpressionFunction(GenericFunction):
    """Turns a Python expression given as a string into a |Function|.

    Some |NumPy| arithmetic functions like 'sin', 'log', 'min' are supported.
    For a full list see the `functions` class attribute.

    .. warning::
       :meth:`eval` is used to evaluate the given expression.
       Using this class with expression strings from untrusted sources will cause
       mayhem and destruction!

    Parameters
    ----------
    expression
        A Python expression of one variable and the `parameters`, given as a `str`.
    dim_domain
        The dimension of the domain.
    parameters
        The |Parameters| the expression accepts.
    values
        Dictionary of additional constants that can be used in `expression`
        with their corresponding value.
    variable
        Name of the input variable in the given expression.
    name
        The name of the function.
    """

    def __init__(self, expression, dim_domain=1, parameters={}, values={}, variable='x', name=None):
        params = parameters.copy()
        params[variable] = dim_domain
        self.expression_obj = parse_expression(expression, parameters=params, values=values)
        super().__init__(self.expression_obj.to_numpy([variable]),
                         dim_domain, self.expression_obj.shape, parameters, name)
        self.__auto_init(locals())

    def __reduce__(self):
        return (ExpressionFunction,
                (self.expression, self.dim_domain, self.parameters, self.values, self.variable,
                 getattr(self, '_name', None)))

    def __str__(self):
        return f'{{x -> {self.expression}}}'


class LincombFunction(Function):
    """A |Function| representing a linear combination of |Functions|.

    The linear coefficients can be provided either as scalars or as
    |ParameterFunctionals|.

    Parameters
    ----------
    functions
        List of |Functions| whose linear combination is formed.
    coefficients
        A list of linear coefficients. A linear coefficient can
        either be a fixed number or a |ParameterFunctional|.
    name
        Name of the function.

    Attributes
    ----------
    functions
    coefficients
    """

    def __init__(self, functions, coefficients, name=None):
        assert len(functions) > 0
        assert len(functions) == len(coefficients)
        assert all(isinstance(f, Function) for f in functions)
        assert all(isinstance(c, (ParameterFunctional, Number)) for c in coefficients)
        assert all(f.dim_domain == functions[0].dim_domain for f in functions[1:])
        assert all(f.shape_range == functions[0].shape_range for f in functions[1:])
        functions = tuple(functions)
        coefficients = tuple(coefficients)

        self.__auto_init(locals())
        self.dim_domain = functions[0].dim_domain
        self.shape_range = functions[0].shape_range

    def evaluate_coefficients(self, mu):
        """Compute the linear coefficients for given |parameter values| `mu`."""
        assert self.parameters.assert_compatible(mu)
        return np.array([c.evaluate(mu) if hasattr(c, 'evaluate') else c for c in self.coefficients])

    def evaluate(self, x, mu=None):
        assert self.parameters.assert_compatible(mu)
        coeffs = self.evaluate_coefficients(mu)
        return sum(c * f(x, mu) for c, f in zip(coeffs, self.functions))


class ProductFunction(Function):
    """A |Function| representing a product of |Functions|.

    Parameters
    ----------
    functions
        List of |Functions| whose product is formed.
    name
        Name of the function.

    Attributes
    ----------
    functions
    """

    def __init__(self, functions, name=None):
        assert len(functions) > 0
        assert all(isinstance(f, Function) for f in functions)
        assert all(f.dim_domain == functions[0].dim_domain for f in functions[1:])
        assert all(f.shape_range == functions[0].shape_range for f in functions[1:])
        self.__auto_init(locals())
        self.dim_domain = functions[0].dim_domain
        self.shape_range = functions[0].shape_range

    def evaluate(self, x, mu=None):
        assert self.parameters.assert_compatible(mu)
        return np.prod([f(x, mu) for f in self.functions], axis=0)


class BitmapFunction(Function):
    """Define a 2D |Function| via a grayscale image.

    Parameters
    ----------
    filename
        Path of the image representing the function.
    bounding_box
        Lower left and upper right coordinates of the domain of the function.
    range
        A pixel of value p is mapped to `(p / 255.) * range[1] + range[0]`.
    """

    dim_domain = 2
    shape_range = ()

    def __init__(self, filename, bounding_box=None, range=None):
        bounding_box = bounding_box or [[0., 0.], [1., 1.]]
        range = range or [0., 1.]
        try:
            from PIL import Image
        except ImportError:
            raise ImportError("PIL is needed for loading images. Try 'pip install pillow'")
        img = Image.open(filename)
        if not img.mode == "L":
            self.logger.warning("Image " + filename + " not in grayscale mode. Converting to grayscale.")
            img = img.convert('L')
        self.__auto_init(locals())
        self.bitmap = np.array(img).T[:, ::-1]
        self.lower_left = np.array(bounding_box[0])
        self.size = np.array(bounding_box[1] - self.lower_left)

    def evaluate(self, x, mu=None):
        indices = np.maximum(np.floor((x - self.lower_left) * np.array(self.bitmap.shape) / self.size).astype(int), 0)
        F = (self.bitmap[np.minimum(indices[..., 0], self.bitmap.shape[0] - 1),
                         np.minimum(indices[..., 1], self.bitmap.shape[1] - 1)]
             * ((self.range[1] - self.range[0]) / 255.)
             + self.range[0])
        return F


class EmpiricalInterpolatedFunction(LincombFunction):
    """Empirically interpolated |Function|.

    Instantiated by :func:`~pymor.algorithm.ei.interpolate_function`.

    Parameters
    ----------
    function
        The |Function| to interpolate.
    interpolation_points
        |NumPy array| containing the coordinates at which the function
        is interpolated. Typically `X[dofs]` where `X` is the array of
        evaluation points used for snapshot data generation and `dofs`
        is returned by :func:`~pymor.algorithm.ei.ei_greedy`.
    interpolation_matrix
        The interpolation matrix corresponding to the selected interpolation
        basis vectors and `interpolation_points` as returned by
        :func:`pymor.algorithms.ei.ei_greedy`.
    triangular
        Whether or not `interpolation_matrix` is lower triangular with unit
        diagonal.
    snapshot_mus
        List of |parameter values| for which the snapshot data for
        :func:`pymor.algorithms.ei.ei_greedy` has been computed.
    snapshot_coefficients
        Matrix of linear coefficients s.t. the i-th interpolation basis vector
        is given by a linear combination of the functions corresponding to
        `snapshot_mus` with the i-th row of `snapshot_coefficients` as
        coefficients. Returned by :func:`~pymor.algorithm.ei.ei_greedy` as
        `data['coefficients']`.
    evaluation_points
        Optional |NumPy array| of coordinates at which the function has been
        evaluated to obtain the snapshot data. If the same evaluation points
        are used to evaluate :class:`EmpiricalInterpolatedFunction`, then
        re-evaluation of the snapshot data at the evaluation points can be
        avoided, when this argument is specified together with `basis_evaluations`.
    basis_evaluations
        Optional |Numpy array| of evaluations of the interpolation basis at
        `evaluation_points`. Corresponds to the `basis` return value of
        :func:`~pymor.algorithms.ei.ei_greedy`.
    """

    def __init__(self, function, interpolation_points, interpolation_matrix, triangular,
                 snapshot_mus, snapshot_coefficients,
                 evaluation_points=None, basis_evaluations=None,
                 name=None):

        assert isinstance(function, Function)
        if len(function.shape_range) > 0:
            raise NotImplementedError
        assert isinstance(interpolation_points, np.ndarray) and interpolation_points.ndim == 2 and \
            interpolation_points.shape[1] == function.dim_domain
        assert isinstance(interpolation_matrix, np.ndarray) and \
            interpolation_matrix.shape == (len(interpolation_points),) * 2
        assert all(isinstance(mu, Mu) for mu in snapshot_mus)
        assert isinstance(snapshot_coefficients, np.ndarray) and \
            snapshot_coefficients.shape == (len(interpolation_points), len(snapshot_mus))
        assert (evaluation_points is None) == (basis_evaluations is None)
        assert evaluation_points is None or isinstance(evaluation_points, np.ndarray) and \
            evaluation_points.ndim == 2 and evaluation_points.shape[1] == function.dim_domain
        assert basis_evaluations is None or isinstance(basis_evaluations, np.ndarray) and \
            basis_evaluations.shape == (len(interpolation_points), len(evaluation_points)) + function.shape_range

        self.__auto_init(locals())
        functions = [EmpiricalInterpolatedFunctionBasisFunction(self, i) for i in range(len(interpolation_points))]
        coefficients = [EmpiricalInterpolatedFunctionFunctional(self, i) for i in range(len(interpolation_points))]
        super().__init__(functions, coefficients)
        self._last_evaluation_points, self._last_basis_evaluations = evaluation_points, basis_evaluations
        if self._last_evaluation_points is not None:
            self._last_evaluation_points.flags.writeable = False
            self._last_basis_evaluations.flags.writeable = False
        self._last_mu = 'NONE'

    def _update_coefficients(self, mu):
        if mu is self._last_mu:
            return
        assert self.parameters.assert_compatible(mu)
        self._last_mu = mu
        fx = self.function.evaluate(self.interpolation_points, mu=mu)
        if self.triangular:
            self._last_interpolation_coefficients = solve_triangular(
                self.interpolation_matrix, fx, lower=True, unit_diagonal=True
            )
        else:
            self._last_interpolation_coefficients = solve(self.interpolation_matrix, fx)

    def _update_evaluation_points(self, x):
        if self._last_evaluation_points is not None and np.all(x == self._last_evaluation_points):
            return
        if self._last_evaluation_points is not None:
            self.logger.info('Evaluating function snapshots for new evaluation points.')
        x.flags.writeable = False
        self._last_evaluation_points = x
        snapshot_evaluations = np.array([self.function.evaluate(x, mu=mu) for mu in self.snapshot_mus])
        self._last_basis_evaluations = self.snapshot_coefficients @ snapshot_evaluations.T


class EmpiricalInterpolatedFunctionFunctional(ParameterFunctional):

    def __init__(self, interpolated_function, index):
        self.__auto_init(locals())
        # we explicitly have to set the parameters since interpolation_points isn't initialized yet
        self.parameters = interpolated_function.function.parameters

    def evaluate(self, mu=None):
        self.interpolated_function._update_coefficients(mu)
        return self.interpolated_function._last_interpolation_coefficients[self.index]


class EmpiricalInterpolatedFunctionBasisFunction(Function):

    def __init__(self, interpolated_function, index):
        self.__auto_init(locals())
        self.dim_domain = interpolated_function.function.dim_domain
        self.shape_range = interpolated_function.function.shape_range
        self.parameters = {}

    def evaluate(self, x, mu=None):
        self.interpolated_function._update_evaluation_points(x)
        return self.interpolated_function._last_basis_evaluations[self.index].copy()



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import itertools
import os
import tempfile
from math import exp, factorial, pi, sin
from pathlib import Path

import numpy as np
import pytest
from hypothesis import given

from pymor.core.config import config
from pymor.core.logger import getLogger
from pymor.discretizers.builtin.grids.vtkio import write_vtk
from pymor.discretizers.builtin.quadratures import GaussQuadratures
from pymor.tools import formatsrc, timing
from pymor.tools.deprecated import Deprecated
from pymor.tools.floatcmp import almost_less, float_cmp, float_cmp_all
from pymor.tools.formatsrc import print_source
from pymor.tools.io import safe_temporary_filename, change_to_directory, read_vtkfile
from pymor.vectorarrays.numpy import NumpyVectorSpace
from pymortests.base import runmodule
from pymortests.fixtures.grid import hy_rect_or_tria_grid

logger = getLogger('pymortests.tools')


FUNCTIONS = (('sin(2x pi)', lambda x: sin(2 * x * pi), 0),
             ('e^x', lambda x: exp(x), exp(1) - exp(0)))


def polynomials(max_order):
    for n in range(max_order + 1):
        def f(x):
            return np.power(x, n)

        def deri(k):
            if k > n:
                return lambda _: 0
            return lambda x: (factorial(n) / factorial(n - k)) * np.power(x, n - k)

        integral = (1 / (n + 1))
        yield (n, f, deri, integral)


def test_quadrature_polynomials():
    for n, function, _, integral in polynomials(GaussQuadratures.orders[-1]):
        name = f'x^{n}'
        for order in GaussQuadratures.orders:
            if n > order / 2:
                continue
            Q = GaussQuadratures.iter_quadrature(order)
            ret = sum([function(p) * w for (p, w) in Q])
            assert float_cmp(ret, integral), \
                f'{name} integral wrong: {integral} vs {ret} (quadrature order {order})'


def test_quadrature_other_functions():
    order = GaussQuadratures.orders[-1]
    for name, function, integral in FUNCTIONS:
        Q = GaussQuadratures.iter_quadrature(order)
        ret = sum([function(p) * w for (p, w) in Q])
        assert float_cmp(ret, integral), \
            f'{name} integral wrong: {integral} vs {ret} (quadrature order {order})'


def test_quadrature_weights():
    for order in GaussQuadratures.orders:
        _, W = GaussQuadratures.quadrature(order)
        assert float_cmp(sum(W), 1)


def test_quadrature_points():
    for order in GaussQuadratures.orders:
        P, _ = GaussQuadratures.quadrature(order)
        assert float_cmp_all(P, np.sort(P))
        assert 0.0 < P[0]
        assert P[-1] < 1.0


def test_float_cmp():
    tol_range = [0.0, 1e-8, 1]
    nan = float('nan')
    inf = float('inf')
    for (rtol, atol) in itertools.product(tol_range, tol_range):
        msg = f'rtol: {rtol} | atol {atol}'
        assert float_cmp(0., 0., rtol, atol), msg
        assert float_cmp(-0., -0., rtol, atol), msg
        assert float_cmp(-1., -1., rtol, atol), msg
        assert float_cmp(0., -0., rtol, atol), msg
        assert not float_cmp(2., -2., rtol, atol), msg

        assert not float_cmp(nan, nan, rtol, atol), msg
        assert nan != nan
        assert not (nan == nan)
        assert not float_cmp(-nan, nan, rtol, atol), msg

        assert not float_cmp(inf, inf, rtol, atol), msg
        assert not (inf != inf)
        assert inf == inf
        if rtol > 0:
            assert float_cmp(-inf, inf, rtol, atol), msg
        else:
            assert not float_cmp(-inf, inf, rtol, atol), msg


def test_almost_less():
    tol_range = [0.0, 1e-8, 1]
    inf = float('inf')
    for (rtol, atol) in itertools.product(tol_range, tol_range):
        msg = f'rtol: {rtol} | atol {atol}'
        assert almost_less(0., 1, rtol, atol), msg
        assert almost_less(-1., -0., rtol, atol), msg
        assert almost_less(-1, -0.9, rtol, atol), msg
        assert almost_less(0., 1, rtol, atol), msg
        assert almost_less(-1., 1., rtol, atol), msg
        assert almost_less(atol, 0., rtol, atol), msg
        assert (rtol == 0.0 and not almost_less(0., inf, rtol, atol), msg) or \
               almost_less(0., inf, rtol, atol), msg


@pytest.mark.skipif(not config.HAVE_VTKIO, reason='VTKIO support libraries missing')
@given(hy_rect_or_tria_grid)
def test_vtkio(grid):
    steps = 4
    for codim, data in enumerate((NumpyVectorSpace.from_numpy(np.ones((steps, grid.size(c))))
                                  for c in range(grid.dim+1))):
        with safe_temporary_filename('wb') as out_name:
            fn = write_vtk(grid, data, out_name, codim=codim)
            meshes = read_vtkfile(fn)
            assert len(meshes) == len(data)
            assert all((a is not None and b is not None for a, b in meshes))


def testTimingContext():
    with timing.Timer('busywait', logger):
        timing.busywait(100)
    with timing.Timer('defaultlog'):
        timing.busywait(100)


def testTimingDecorator():

    @timing.Timer('busywait_decorator', logger)
    def wait():
        timing.busywait(1000)

    wait()


def testTiming():
    timer = timing.Timer('busywait', logger)
    timer.start()
    timing.busywait(1000)
    timer.stop()
    logger.info('plain timing took %s seconds', timer.dt)


def testDeprecated():
    @Deprecated('use other stuff instead')
    def deprecated_function():
        pass
    # Cause all warnings to always be triggered.
    import warnings
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        # Trigger a warning.
        deprecated_function()
        # Verify some things
        assert len(w) == 1
        assert issubclass(w[-1].category, DeprecationWarning)
        assert "DeprecationWarning" in str(w[-1].message)


def test_formatsrc():
    obj = formatsrc.format_source
    formatsrc.format_source(obj)
    print_source(obj)


def test_formatsrc_nopygments(monkeypatch):
    try:
        from pygments import highlight  # noqa: F401
        monkeypatch.delattr('pygments.highlight')
    except ImportError:
        pass
    test_formatsrc()


def test_load_matrix(loadable_matrices):
    from pymor.tools.io import load_matrix
    for m in loadable_matrices:
        if m.suffix == '.npz':
            load_matrix(m, key='arr_0')
        else:
            load_matrix(m)


@pytest.mark.parametrize('ext', ['.mat', '.mtx', '.mtz.gz', '.npy', '.npz', '.txt'])
def test_save_load_matrix(ext):
    import filecmp
    from pymor.tools.io import load_matrix, save_matrix
    A = np.eye(2)
    with tempfile.TemporaryDirectory() as tmpdirname:
        path = os.path.join(tmpdirname, 'matrix' + ext)
        key = None
        if ext == '.mat':
            key = 'A'
        save_matrix(path, A, key)
        B = load_matrix(path, key)
        assert np.all(A == B)
        path2 = os.path.join(tmpdirname, 'matrix2' + ext)
        save_matrix(path2, A, key)
        # .mat save a timestamp, so full file cmp os too flaky
        if ext not in ('.mtz.gz', '.mat'):
            assert filecmp.cmp(path, path2)


def test_cwd_ctx_manager():
    def _cwd():
        return Path(os.getcwd()).resolve()

    original_cwd = _cwd()
    target = Path(tempfile.gettempdir()).resolve()
    with change_to_directory(target) as result:
        assert result is None
        assert _cwd() == target
    assert _cwd() == original_cwd


def test_deprecated_tmp():
    """This test should be removed alongside SafeTemporaryFileName after the next release"""
    from pymor.tools.io import SafeTemporaryFileName
    with SafeTemporaryFileName() as fn:
        open(fn, 'wt').write('foo')


if __name__ == "__main__":
    runmodule(filename=__file__)



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)


import numpy as np
from scipy.linalg import solve, solve_continuous_lyapunov, solve_continuous_are
from scipy.sparse.linalg import bicgstab, spsolve, splu, spilu, lgmres, lsqr, LinearOperator

from pymor.algorithms.lyapunov import _solve_lyap_lrcf_check_args, _solve_lyap_dense_check_args, _chol
from pymor.algorithms.riccati import _solve_ricc_check_args, _solve_ricc_dense_check_args
from pymor.algorithms.genericsolvers import _parse_options
from pymor.algorithms.to_matrix import to_matrix
from pymor.core.config import config
from pymor.core.defaults import defaults
from pymor.core.exceptions import InversionError
from pymor.operators.numpy import NumpyMatrixOperator


@defaults('bicgstab_tol', 'bicgstab_maxiter', 'spilu_drop_tol',
          'spilu_fill_factor', 'spilu_drop_rule', 'spilu_permc_spec', 'spsolve_permc_spec',
          'spsolve_keep_factorization',
          'lgmres_tol', 'lgmres_maxiter', 'lgmres_inner_m', 'lgmres_outer_k', 'least_squares_lsmr_damp',
          'least_squares_lsmr_atol', 'least_squares_lsmr_btol', 'least_squares_lsmr_conlim',
          'least_squares_lsmr_maxiter', 'least_squares_lsmr_show', 'least_squares_lsqr_atol',
          'least_squares_lsqr_btol', 'least_squares_lsqr_conlim', 'least_squares_lsqr_iter_lim',
          'least_squares_lsqr_show')
def solver_options(bicgstab_tol=1e-15,
                   bicgstab_maxiter=None,
                   spilu_drop_tol=1e-4,
                   spilu_fill_factor=10,
                   spilu_drop_rule=None,
                   spilu_permc_spec='COLAMD',
                   spsolve_permc_spec='COLAMD',
                   spsolve_keep_factorization=True,
                   lgmres_tol=1e-5,
                   lgmres_maxiter=1000,
                   lgmres_inner_m=39,
                   lgmres_outer_k=3,
                   least_squares_lsmr_damp=0.0,
                   least_squares_lsmr_atol=1e-6,
                   least_squares_lsmr_btol=1e-6,
                   least_squares_lsmr_conlim=1e8,
                   least_squares_lsmr_maxiter=None,
                   least_squares_lsmr_show=False,
                   least_squares_lsqr_damp=0.0,
                   least_squares_lsqr_atol=1e-6,
                   least_squares_lsqr_btol=1e-6,
                   least_squares_lsqr_conlim=1e8,
                   least_squares_lsqr_iter_lim=None,
                   least_squares_lsqr_show=False):
    """Returns available solvers with default |solver_options| for the SciPy backend.

    Parameters
    ----------
    bicgstab_tol
        See :func:`scipy.sparse.linalg.bicgstab`.
    bicgstab_maxiter
        See :func:`scipy.sparse.linalg.bicgstab`.
    spilu_drop_tol
        See :func:`scipy.sparse.linalg.spilu`.
    spilu_fill_factor
        See :func:`scipy.sparse.linalg.spilu`.
    spilu_drop_rule
        See :func:`scipy.sparse.linalg.spilu`.
    spilu_permc_spec
        See :func:`scipy.sparse.linalg.spilu`.
    spsolve_permc_spec
        See :func:`scipy.sparse.linalg.spsolve`.
    spsolve_keep_factorization
        See :func:`scipy.sparse.linalg.spsolve`.
    lgmres_tol
        See :func:`scipy.sparse.linalg.lgmres`.
    lgmres_maxiter
        See :func:`scipy.sparse.linalg.lgmres`.
    lgmres_inner_m
        See :func:`scipy.sparse.linalg.lgmres`.
    lgmres_outer_k
        See :func:`scipy.sparse.linalg.lgmres`.
    least_squares_lsmr_damp
        See :func:`scipy.sparse.linalg.lsmr`.
    least_squares_lsmr_atol
        See :func:`scipy.sparse.linalg.lsmr`.
    least_squares_lsmr_btol
        See :func:`scipy.sparse.linalg.lsmr`.
    least_squares_lsmr_conlim
        See :func:`scipy.sparse.linalg.lsmr`.
    least_squares_lsmr_maxiter
        See :func:`scipy.sparse.linalg.lsmr`.
    least_squares_lsmr_show
        See :func:`scipy.sparse.linalg.lsmr`.
    least_squares_lsqr_damp
        See :func:`scipy.sparse.linalg.lsqr`.
    least_squares_lsqr_atol
        See :func:`scipy.sparse.linalg.lsqr`.
    least_squares_lsqr_btol
        See :func:`scipy.sparse.linalg.lsqr`.
    least_squares_lsqr_conlim
        See :func:`scipy.sparse.linalg.lsqr`.
    least_squares_lsqr_iter_lim
        See :func:`scipy.sparse.linalg.lsqr`.
    least_squares_lsqr_show
        See :func:`scipy.sparse.linalg.lsqr`.

    Returns
    -------
    A dict of available solvers with default |solver_options|.
    """
    opts = {'scipy_bicgstab_spilu':     {'type': 'scipy_bicgstab_spilu',
                                         'tol': bicgstab_tol,
                                         'maxiter': bicgstab_maxiter,
                                         'spilu_drop_tol': spilu_drop_tol,
                                         'spilu_fill_factor': spilu_fill_factor,
                                         'spilu_drop_rule': spilu_drop_rule,
                                         'spilu_permc_spec': spilu_permc_spec},
            'scipy_bicgstab':           {'type': 'scipy_bicgstab',
                                         'tol': bicgstab_tol,
                                         'maxiter': bicgstab_maxiter},
            'scipy_spsolve':            {'type': 'scipy_spsolve',
                                         'permc_spec': spsolve_permc_spec,
                                         'keep_factorization': spsolve_keep_factorization},
            'scipy_lgmres':             {'type': 'scipy_lgmres',
                                         'tol': lgmres_tol,
                                         'maxiter': lgmres_maxiter,
                                         'inner_m': lgmres_inner_m,
                                         'outer_k': lgmres_outer_k},
            'scipy_least_squares_lsqr': {'type': 'scipy_least_squares_lsqr',
                                         'damp': least_squares_lsqr_damp,
                                         'atol': least_squares_lsqr_atol,
                                         'btol': least_squares_lsqr_btol,
                                         'conlim': least_squares_lsqr_conlim,
                                         'iter_lim': least_squares_lsqr_iter_lim,
                                         'show': least_squares_lsqr_show}}

    if config.HAVE_SCIPY_LSMR:
        opts['scipy_least_squares_lsmr'] = {'type': 'scipy_least_squares_lsmr',
                                            'damp': least_squares_lsmr_damp,
                                            'atol': least_squares_lsmr_atol,
                                            'btol': least_squares_lsmr_btol,
                                            'conlim': least_squares_lsmr_conlim,
                                            'maxiter': least_squares_lsmr_maxiter,
                                            'show': least_squares_lsmr_show}

    return opts


@defaults('check_finite', 'default_solver', 'default_least_squares_solver')
def apply_inverse(op, V, initial_guess=None, options=None, least_squares=False, check_finite=True,
                  default_solver='scipy_spsolve', default_least_squares_solver='scipy_least_squares_lsmr'):
    """Solve linear equation system.

    Applies the inverse of `op` to the vectors in `V` using SciPy.

    Parameters
    ----------
    op
        The linear, non-parametric |Operator| to invert.
    V
        |VectorArray| of right-hand sides for the equation system.
    initial_guess
        |VectorArray| with the same length as `V` containing initial guesses
        for the solution.  Some implementations of `apply_inverse` may
        ignore this parameter.  If `None` a solver-dependent default is used.
    options
        The |solver_options| to use (see :func:`solver_options`).
    least_squares
        If `True`, return least squares solution.
    check_finite
        Test if solution only contains finite values.
    default_solver
        Default solver to use (scipy_spsolve, scipy_bicgstab, scipy_bicgstab_spilu,
        scipy_lgmres, scipy_least_squares_lsmr, scipy_least_squares_lsqr).
    default_least_squares_solver
        Default solver to use for least squares problems (scipy_least_squares_lsmr,
        scipy_least_squares_lsqr).

    Returns
    -------
    |VectorArray| of the solution vectors.
    """
    assert V in op.range
    assert initial_guess is None or initial_guess in op.source and len(initial_guess) == len(V)

    if isinstance(op, NumpyMatrixOperator):
        matrix = op.matrix
    else:
        from pymor.algorithms.to_matrix import to_matrix
        matrix = to_matrix(op)

    options = _parse_options(options, solver_options(), default_solver, default_least_squares_solver, least_squares)

    V = V.to_numpy()
    initial_guess = initial_guess.to_numpy() if initial_guess is not None else None
    promoted_type = np.promote_types(matrix.dtype, V.dtype)
    R = np.empty((len(V), matrix.shape[1]), dtype=promoted_type)

    if options['type'] == 'scipy_bicgstab':
        for i, VV in enumerate(V):
            R[i], info = bicgstab(matrix, VV, initial_guess[i] if initial_guess is not None else None,
                                  tol=options['tol'], maxiter=options['maxiter'])
            if info != 0:
                if info > 0:
                    raise InversionError(f'bicgstab failed to converge after {info} iterations')
                else:
                    raise InversionError('bicgstab failed with error code {} (illegal input or breakdown)'.
                                         format(info))
    elif options['type'] == 'scipy_bicgstab_spilu':
        ilu = spilu(matrix, drop_tol=options['spilu_drop_tol'], fill_factor=options['spilu_fill_factor'],
                    drop_rule=options['spilu_drop_rule'], permc_spec=options['spilu_permc_spec'])
        precond = LinearOperator(matrix.shape, ilu.solve)
        for i, VV in enumerate(V):
            R[i], info = bicgstab(matrix, VV, initial_guess[i] if initial_guess is not None else None,
                                  tol=options['tol'], maxiter=options['maxiter'], M=precond)
            if info != 0:
                if info > 0:
                    raise InversionError(f'bicgstab failed to converge after {info} iterations')
                else:
                    raise InversionError('bicgstab failed with error code {} (illegal input or breakdown)'.
                                         format(info))
    elif options['type'] == 'scipy_spsolve':
        try:
            # maybe remove unusable factorization:
            if hasattr(matrix, 'factorization'):
                fdtype = matrix.factorizationdtype
                if not np.can_cast(V.dtype, fdtype, casting='safe'):
                    del matrix.factorization

            if hasattr(matrix, 'factorization'):
                # we may use a complex factorization of a real matrix to
                # apply it to a real vector. In that case, we downcast
                # the result here, removing the imaginary part,
                # which should be zero.
                R = matrix.factorization.solve(V.T).T.astype(promoted_type, copy=False)
            elif options['keep_factorization']:
                # the matrix is always converted to the promoted type.
                # if matrix.dtype == promoted_type, this is a no_op
                matrix.factorization = splu(matrix_astype_nocopy(matrix.tocsc(), promoted_type),
                                            permc_spec=options['permc_spec'])
                matrix.factorizationdtype = promoted_type
                R = matrix.factorization.solve(V.T).T
            else:
                # the matrix is always converted to the promoted type.
                # if matrix.dtype == promoted_type, this is a no_op
                R = spsolve(matrix_astype_nocopy(matrix, promoted_type), V.T, permc_spec=options['permc_spec']).T
        except RuntimeError as e:
            raise InversionError(e)
    elif options['type'] == 'scipy_lgmres':
        for i, VV in enumerate(V):
            R[i], info = lgmres(matrix, VV, initial_guess[i] if initial_guess is not None else None,
                                tol=options['tol'],
                                atol=options['tol'],
                                maxiter=options['maxiter'],
                                inner_m=options['inner_m'],
                                outer_k=options['outer_k'])
            if info > 0:
                raise InversionError(f'lgmres failed to converge after {info} iterations')
            assert info == 0
    elif options['type'] == 'scipy_least_squares_lsmr':
        from scipy.sparse.linalg import lsmr
        for i, VV in enumerate(V):
            R[i], info, itn, _, _, _, _, _ = lsmr(matrix, VV,
                                                  damp=options['damp'],
                                                  atol=options['atol'],
                                                  btol=options['btol'],
                                                  conlim=options['conlim'],
                                                  maxiter=options['maxiter'],
                                                  show=options['show'],
                                                  x0=initial_guess[i] if initial_guess is not None else None)
            assert 0 <= info <= 7
            if info == 7:
                raise InversionError(f'lsmr failed to converge after {itn} iterations')
    elif options['type'] == 'scipy_least_squares_lsqr':
        for i, VV in enumerate(V):
            R[i], info, itn, _, _, _, _, _, _, _ = lsqr(matrix, VV,
                                                        damp=options['damp'],
                                                        atol=options['atol'],
                                                        btol=options['btol'],
                                                        conlim=options['conlim'],
                                                        iter_lim=options['iter_lim'],
                                                        show=options['show'],
                                                        x0=initial_guess[i] if initial_guess is not None else None)
            assert 0 <= info <= 7
            if info == 7:
                raise InversionError(f'lsmr failed to converge after {itn} iterations')
    else:
        raise ValueError('Unknown solver type')

    if check_finite:
        if not np.isfinite(np.sum(R)):
            raise InversionError('Result contains non-finite values')

    return op.source.from_numpy(R)


# unfortunately, this is necessary, as scipy does not
# forward the copy=False argument in its csc_matrix.astype function
def matrix_astype_nocopy(matrix, dtype):
    if matrix.dtype == dtype:
        return matrix
    else:
        return matrix.astype(dtype)


def lyap_lrcf_solver_options():
    """Return available Lyapunov solvers with default options for the SciPy backend.

    Returns
    -------
    A dict of available solvers with default solver options.
    """
    return {'scipy': {'type': 'scipy'}}


def solve_lyap_lrcf(A, E, B, trans=False, options=None):
    """Compute an approximate low-rank solution of a Lyapunov equation.

    See :func:`pymor.algorithms.lyapunov.solve_lyap_lrcf` for a general
    description.

    This function uses `scipy.linalg.solve_continuous_lyapunov`, which
    is a dense solver for Lyapunov equations with E=I.
    Therefore, we assume A and E can be converted to |NumPy arrays|
    using :func:`~pymor.algorithms.to_matrix.to_matrix` and that
    `B.to_numpy` is implemented.

    .. note::
        If E is not `None`, the problem will be reduced to a standard
        continuous-time algebraic Lyapunov equation by inverting E.

    Parameters
    ----------
    A
        The non-parametric |Operator| A.
    E
        The non-parametric |Operator| E or `None`.
    B
        The operator B as a |VectorArray| from `A.source`.
    trans
        Whether the first |Operator| in the Lyapunov equation is
        transposed.
    options
        The solver options to use (see
        :func:`lyap_lrcf_solver_options`).

    Returns
    -------
    Z
        Low-rank Cholesky factor of the Lyapunov equation solution,
        |VectorArray| from `A.source`.
    """
    _solve_lyap_lrcf_check_args(A, E, B, trans)
    options = _parse_options(options, lyap_lrcf_solver_options(), 'scipy', None, False)

    X = solve_lyap_dense(to_matrix(A, format='dense'),
                         to_matrix(E, format='dense') if E else None,
                         B.to_numpy().T if not trans else B.to_numpy(),
                         trans=trans, options=options)
    return A.source.from_numpy(_chol(X).T)


def lyap_dense_solver_options():
    """Return available dense Lyapunov solvers with default options for the SciPy backend.

    Returns
    -------
    A dict of available solvers with default solver options.
    """
    return {'scipy': {'type': 'scipy'}}


def solve_lyap_dense(A, E, B, trans=False, options=None):
    """Compute the solution of a Lyapunov equation.

    See :func:`pymor.algorithms.lyapunov.solve_lyap_dense` for a
    general description.

    This function uses `scipy.linalg.solve_continuous_lyapunov`, which
    is a dense solver for Lyapunov equations with E=I.

    .. note::
        If E is not `None`, the problem will be reduced to a standard
        continuous-time algebraic Lyapunov equation by inverting E.

    Parameters
    ----------
    A
        The matrix A as a 2D |NumPy array|.
    E
        The matrix E as a 2D |NumPy array| or `None`.
    B
        The matrix B as a 2D |NumPy array|.
    trans
        Whether the first operator in the Lyapunov equation is
        transposed.
    options
        The solver options to use (see
        :func:`lyap_dense_solver_options`).

    Returns
    -------
    X
        Lyapunov equation solution as a |NumPy array|.
    """
    _solve_lyap_dense_check_args(A, E, B, trans)
    options = _parse_options(options, lyap_dense_solver_options(), 'scipy', None, False)

    if options['type'] == 'scipy':
        if E is not None:
            A = solve(E, A) if not trans else solve(E.T, A.T).T
            B = solve(E, B) if not trans else solve(E.T, B.T).T
        if trans:
            A = A.T
            B = B.T
        X = solve_continuous_lyapunov(A, -B.dot(B.T))
    else:
        raise ValueError(f"Unexpected Lyapunov equation solver ({options['type']}).")

    return X


def ricc_lrcf_solver_options():
    """Return available Riccati solvers with default options for the SciPy backend.

    Returns
    -------
    A dict of available solvers with default solver options.
    """
    return {'scipy': {'type': 'scipy'}}


def solve_ricc_lrcf(A, E, B, C, R=None, trans=False, options=None):
    """Compute an approximate low-rank solution of a Riccati equation.

    See :func:`pymor.algorithms.riccati.solve_ricc_lrcf` for a general
    description.

    This function uses `scipy.linalg.solve_continuous_are`, which
    is a dense solver.
    Therefore, we assume all |Operators| and |VectorArrays| can be
    converted to |NumPy arrays| using
    :func:`~pymor.algorithms.to_matrix.to_matrix` and
    :func:`~pymor.vectorarrays.interface.VectorArray.to_numpy`.

    Parameters
    ----------
    A
        The non-parametric |Operator| A.
    E
        The non-parametric |Operator| E or `None`.
    B
        The operator B as a |VectorArray| from `A.source`.
    C
        The operator C as a |VectorArray| from `A.source`.
    R
        The matrix R as a 2D |NumPy array| or `None`.
    trans
        Whether the first |Operator| in the Riccati equation is
        transposed.
    options
        The solver options to use (see :func:`ricc_lrcf_solver_options`).

    Returns
    -------
    Z
        Low-rank Cholesky factor of the Riccati equation solution,
        |VectorArray| from `A.source`.
    """
    _solve_ricc_check_args(A, E, B, C, R, trans)
    options = _parse_options(options, ricc_lrcf_solver_options(), 'scipy', None, False)
    if options['type'] != 'scipy':
        raise ValueError(f"Unexpected Riccati equation solver ({options['type']}).")

    A_source = A.source
    A = to_matrix(A, format='dense')
    E = to_matrix(E, format='dense') if E else None
    B = B.to_numpy().T
    C = C.to_numpy()
    X = solve_ricc_dense(A, E, B, C, R, trans, options)

    return A_source.from_numpy(_chol(X).T)


def ricc_dense_solver_options():
    """Return available Riccati solvers with default options for the SciPy backend.

    Returns
    -------
    A dict of available solvers with default solver options.
    """
    return {'scipy': {'type': 'scipy'}}


def solve_ricc_dense(A, E, B, C, R=None, trans=False, options=None):
    """Compute the solution of a Riccati equation.

    See :func:`pymor.algorithms.riccati.solve_ricc_dense` for a general
    description.

    This function uses `scipy.linalg.solve_continuous_are`, which
    is a dense solver.

    Parameters
    ----------
    A
        The matrix A as a 2D |NumPy array|.
    E
        The matrix E as a 2D |NumPy array| or `None`.
    B
        The matrix B as a 2D |NumPy array|.
    C
        The matrix C as a 2D |NumPy array|.
    R
        The matrix R as a 2D |NumPy array| or `None`.
    trans
        Whether the first operator in the Riccati equation is
        transposed.
    options
        The solver options to use (see
        :func:`ricc_dense_solver_options`).

    Returns
    -------
    X
        Riccati equation solution as a |NumPy array|.
    """
    _solve_ricc_dense_check_args(A, E, B, C, R, trans)
    options = _parse_options(options, ricc_dense_solver_options(), 'scipy', None, False)
    if options['type'] != 'scipy':
        raise ValueError(f"Unexpected Riccati equation solver ({options['type']}).")

    if R is None:
        R = np.eye(C.shape[0] if not trans else B.shape[1])
    if not trans:
        if E is not None:
            E = E.T
        X = solve_continuous_are(A.T, C.T, B.dot(B.T), R, E)
    else:
        X = solve_continuous_are(A, B, C.T.dot(C), R, E)

    return X


def pos_ricc_lrcf_solver_options():
    """Return available positive Riccati solvers with default options for the SciPy backend.

    Returns
    -------
    A dict of available solvers with default solver options.
    """
    return {'scipy': {'type': 'scipy'}}


def solve_pos_ricc_lrcf(A, E, B, C, R=None, trans=False, options=None):
    """Compute an approximate low-rank solution of a positive Riccati equation.

    See :func:`pymor.algorithms.riccati.solve_pos_ricc_lrcf` for a
    general description.

    This function uses `scipy.linalg.solve_continuous_are`, which
    is a dense solver.
    Therefore, we assume all |Operators| and |VectorArrays| can be
    converted to |NumPy arrays| using
    :func:`~pymor.algorithms.to_matrix.to_matrix` and
    :func:`~pymor.vectorarrays.interface.VectorArray.to_numpy`.

    Parameters
    ----------
    A
        The non-parametric |Operator| A.
    E
        The non-parametric |Operator| E or `None`.
    B
        The operator B as a |VectorArray| from `A.source`.
    C
        The operator C as a |VectorArray| from `A.source`.
    R
        The matrix R as a 2D |NumPy array| or `None`.
    trans
        Whether the first |Operator| in the positive Riccati equation is
        transposed.
    options
        The solver options to use (see
        :func:`pos_ricc_lrcf_solver_options`).

    Returns
    -------
    Z
        Low-rank Cholesky factor of the positive Riccati equation
        solution, |VectorArray| from `A.source`.
    """
    _solve_ricc_check_args(A, E, B, C, R, trans)
    options = _parse_options(options, pos_ricc_lrcf_solver_options(), 'scipy', None, False)
    if options['type'] != 'scipy':
        raise ValueError(f"Unexpected positive Riccati equation solver ({options['type']}).")

    if R is None:
        R = np.eye(len(C) if not trans else len(B))
    return solve_ricc_lrcf(A, E, B, C, -R, trans, options)



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

from pymor.core.config import config


if config.HAVE_PYMESS:
    import numpy as np
    import scipy.linalg as spla
    import pymess

    from pymor.algorithms.genericsolvers import _parse_options
    from pymor.algorithms.lyapunov import (mat_eqn_sparse_min_size, _solve_lyap_lrcf_check_args,
                                           _solve_lyap_dense_check_args, _chol)
    from pymor.algorithms.to_matrix import to_matrix
    from pymor.bindings.scipy import _solve_ricc_check_args
    from pymor.core.defaults import defaults
    from pymor.core.logger import getLogger
    from pymor.operators.constructions import IdentityOperator

    @defaults('adi_maxit', 'adi_memory_usage', 'adi_output', 'adi_rel_change_tol', 'adi_res2_tol', 'adi_res2c_tol',
              'adi_shifts_arp_m', 'adi_shifts_arp_p', 'adi_shifts_b0', 'adi_shifts_l0', 'adi_shifts_p',
              'adi_shifts_paratype')
    def lradi_solver_options(adi_maxit=500,
                             adi_memory_usage=pymess.MESS_MEMORY_MID,
                             adi_output=1,
                             adi_rel_change_tol=1e-10,
                             adi_res2_tol=1e-10,
                             adi_res2c_tol=1e-11,
                             adi_shifts_arp_m=32,
                             adi_shifts_arp_p=48,
                             adi_shifts_b0=None,
                             adi_shifts_l0=16,
                             adi_shifts_p=None,
                             adi_shifts_paratype=pymess.MESS_LRCFADI_PARA_ADAPTIVE_Z):
        """Return available adi solver options with default values for the pymess backend.

        Parameters
        ----------
        adi_maxit
            See `pymess.OptionsAdi`.
        adi_memory_usage
            See `pymess.OptionsAdi`.
        adi_output
            See `pymess.OptionsAdi`.
        adi_rel_change_tol
            See `pymess.OptionsAdi`.
        adi_res2_tol
            See `pymess.OptionsAdi`.
        adi_res2c_tol
            See `pymess.OptionsAdi`.
        adi_shifts_arp_m
            See `pymess.OptionsAdiShifts`.
        adi_shifts_arp_p
            See `pymess.OptionsAdiShifts`.
        adi_shifts_b0
            See `pymess.OptionsAdiShifts`.
        adi_shifts_l0
            See `pymess.OptionsAdiShifts`.
        adi_shifts_p
            See `pymess.OptionsAdiShifts`.
        adi_shifts_paratype
            See `pymess.OptionsAdiShifts`.

        Returns
        -------
        A dict of available solvers with default solver options.
        """
        lradi_opts = pymess.Options()
        lradi_opts.adi.maxit = adi_maxit
        lradi_opts.adi.memory_usage = adi_memory_usage
        lradi_opts.adi.output = adi_output
        lradi_opts.adi.rel_change_tol = adi_rel_change_tol
        lradi_opts.adi.res2_tol = adi_res2_tol
        lradi_opts.adi.res2c_tol = adi_res2c_tol
        lradi_opts.adi.shifts.arp_m = adi_shifts_arp_m
        lradi_opts.adi.shifts.arp_p = adi_shifts_arp_p
        lradi_opts.adi.shifts.b0 = adi_shifts_b0
        lradi_opts.adi.shifts.l0 = adi_shifts_l0
        lradi_opts.adi.shifts.p = adi_shifts_p
        lradi_opts.adi.shifts.paratype = adi_shifts_paratype
        return lradi_opts

    def lyap_lrcf_solver_options():
        """Return available Lyapunov solvers with default options for the pymess backend.

        Also see :func:`lradi_solver_options`.

        Returns
        -------
        A dict of available solvers with default solver options.
        """
        return {'pymess_glyap': {'type': 'pymess_glyap'},
                'pymess_lradi': {'type': 'pymess_lradi',
                                 'opts': lradi_solver_options()}}

    @defaults('default_solver')
    def solve_lyap_lrcf(A, E, B, trans=False, options=None, default_solver=None):
        """Compute an approximate low-rank solution of a Lyapunov equation.

        See :func:`pymor.algorithms.lyapunov.solve_lyap_lrcf` for a
        general description.

        This function uses `pymess.glyap` and `pymess.lradi`.
        For both methods,
        :meth:`~pymor.vectorarrays.interface.VectorArray.to_numpy`
        and
        :meth:`~pymor.vectorarrays.interface.VectorSpace.from_numpy`
        need to be implemented for `A.source`.
        Additionally, since `glyap` is a dense solver, it expects
        :func:`~pymor.algorithms.to_matrix.to_matrix` to work for A and
        E.

        If the solver is not specified using the options or
        default_solver arguments, `glyap` is used for small problems
        (smaller than defined with
        :func:`~pymor.algorithms.lyapunov.mat_eqn_sparse_min_size`) and
        `lradi` for large problems.

        Parameters
        ----------
        A
            The non-parametric |Operator| A.
        E
            The non-parametric |Operator| E or `None`.
        B
            The operator B as a |VectorArray| from `A.source`.
        trans
            Whether the first |Operator| in the Lyapunov equation is
            transposed.
        options
            The solver options to use (see
            :func:`lyap_lrcf_solver_options`).
        default_solver
            Default solver to use (pymess_lradi, pymess_glyap).
            If `None`, choose solver depending on the dimension of A.

        Returns
        -------
        Z
            Low-rank Cholesky factor of the Lyapunov equation solution,
            |VectorArray| from `A.source`.
        """
        _solve_lyap_lrcf_check_args(A, E, B, trans)
        if default_solver is None:
            default_solver = 'pymess_lradi' if A.source.dim >= mat_eqn_sparse_min_size() else 'pymess_glyap'
        options = _parse_options(options, lyap_lrcf_solver_options(), default_solver, None, False)

        if options['type'] == 'pymess_glyap':
            X = solve_lyap_dense(to_matrix(A, format='dense'),
                                 to_matrix(E, format='dense') if E else None,
                                 B.to_numpy().T if not trans else B.to_numpy(),
                                 trans=trans, options=options)
            Z = _chol(X)
        elif options['type'] == 'pymess_lradi':
            opts = options['opts']
            opts.type = pymess.MESS_OP_NONE if not trans else pymess.MESS_OP_TRANSPOSE
            eqn = LyapunovEquation(opts, A, E, B)
            Z, status = pymess.lradi(eqn, opts)
            relres = status.res2_norm / status.res2_0
            if relres > opts.adi.res2_tol:
                logger = getLogger('pymor.bindings.pymess.solve_lyap_lrcf')
                logger.warning(f'Desired relative residual tolerance was not achieved '
                               f'({relres:e} > {opts.adi.res2_tol:e}).')
        else:
            raise ValueError(f'Unexpected Lyapunov equation solver ({options["type"]}).')

        return A.source.from_numpy(Z.T)

    def lyap_dense_solver_options():
        """Return available Lyapunov solvers with default options for the pymess backend.

        Returns
        -------
        A dict of available solvers with default solver options.
        """
        return {'pymess_glyap': {'type': 'pymess_glyap'}}

    def solve_lyap_dense(A, E, B, trans=False, options=None):
        """Compute the solution of a Lyapunov equation.

        See :func:`pymor.algorithms.lyapunov.solve_lyap_dense` for a
        general description.

        This function uses `pymess.glyap`.

        Parameters
        ----------
        A
            The matrix A as a 2D |NumPy array|.
        E
            The matrix E as a 2D |NumPy array| or `None`.
        B
            The matrix B as a 2D |NumPy array|.
        trans
            Whether the first operator in the Lyapunov equation is
            transposed.
        options
            The solver options to use (see
            :func:`lyap_dense_solver_options`).

        Returns
        -------
        X
            Lyapunov equation solution as a |NumPy array|.
        """
        _solve_lyap_dense_check_args(A, E, B, trans)
        options = _parse_options(options, lyap_lrcf_solver_options(), 'pymess_glyap', None, False)

        if options['type'] == 'pymess_glyap':
            Y = B.dot(B.T) if not trans else B.T.dot(B)
            op = pymess.MESS_OP_NONE if not trans else pymess.MESS_OP_TRANSPOSE
            X = pymess.glyap(A, E, Y, op=op)[0]
            X = np.asarray(X)
        else:
            raise ValueError(f'Unexpected Lyapunov equation solver ({options["type"]}).')

        return X

    @defaults('linesearch', 'maxit', 'absres_tol', 'relres_tol', 'nrm')
    def dense_nm_gmpcare_solver_options(linesearch=False,
                                        maxit=50,
                                        absres_tol=1e-11,
                                        relres_tol=1e-12,
                                        nrm=0):
        """Return available Riccati solvers with default options for the pymess backend.

        Also see :func:`lradi_solver_options`.

        Parameters
        ----------
        linesearch
            See `pymess.dense_nm_gmpcare`.
        maxit
            See `pymess.dense_nm_gmpcare`.
        absres_tol
            See `pymess.dense_nm_gmpcare`.
        relres_tol
            See `pymess.dense_nm_gmpcare`.
        nrm
            See `pymess.dense_nm_gmpcare`.

        Returns
        -------
        A dict of available solvers with default solver options.
        """
        return {'linesearch': linesearch,
                'maxit':      maxit,
                'absres_tol': absres_tol,
                'relres_tol': relres_tol,
                'nrm':        nrm}

    @defaults('newton_gstep', 'newton_k0', 'newton_linesearch', 'newton_maxit', 'newton_output', 'newton_res2_tol',
              'newton_singleshifts')
    def lrnm_solver_options(newton_gstep=0,
                            newton_k0=None,
                            newton_linesearch=0,
                            newton_maxit=30,
                            newton_output=1,
                            newton_res2_tol=1e-10,
                            newton_singleshifts=0):
        """Return available adi solver options with default values for the pymess backend.

        Parameters
        ----------
        newton_gstep
          See `pymess.OptionsNewton`.
        newton_k0
          See `pymess.OptionsNewton`.
        newton_linesearch
          See `pymess.OptionsNewton`.
        newton_maxit
          See `pymess.OptionsNewton`.
        newton_output
          See `pymess.OptionsNewton`.
        newton_res2_tol
          See `pymess.OptionsNewton`.
        newton_singleshifts
          See `pymess.OptionsNewton`.

        Returns
        -------
        A dict of available solvers with default solver options.
        """
        lrnm_opts = lradi_solver_options()
        lrnm_opts.nm.gstep = newton_gstep
        lrnm_opts.nm.k0 = newton_k0
        lrnm_opts.nm.linesearch = newton_linesearch
        lrnm_opts.nm.maxit = newton_maxit
        lrnm_opts.nm.output = newton_output
        lrnm_opts.nm.res2_tol = newton_res2_tol
        lrnm_opts.nm.singleshifts = newton_singleshifts

        return lrnm_opts

    def ricc_lrcf_solver_options():
        """Return available Riccati solvers with default options for the pymess backend.

        Also see :func:`dense_nm_gmpcare_solver_options` and
        :func:`lrnm_solver_options`.

        Returns
        -------
        A dict of available solvers with default solver options.
        """
        return {'pymess_dense_nm_gmpcare': {'type': 'pymess_dense_nm_gmpcare',
                                            'opts': dense_nm_gmpcare_solver_options()},
                'pymess_lrnm':             {'type': 'pymess_lrnm',
                                            'opts': lrnm_solver_options()}}

    @defaults('default_solver')
    def solve_ricc_lrcf(A, E, B, C, R=None, trans=False, options=None, default_solver=None):
        """Compute an approximate low-rank solution of a Riccati equation.

        See :func:`pymor.algorithms.riccati.solve_ricc_lrcf` for a
        general description.

        This function uses `pymess.dense_nm_gmpcare` and `pymess.lrnm`.
        For both methods,
        :meth:`~pymor.vectorarrays.interface.VectorArray.to_numpy`
        and
        :meth:`~pymor.vectorarrays.interface.VectorSpace.from_numpy`
        need to be implemented for `A.source`.
        Additionally, since `dense_nm_gmpcare` is a dense solver, it
        expects :func:`~pymor.algorithms.to_matrix.to_matrix` to work
        for A and E.

        If the solver is not specified using the options or
        default_solver arguments, `dense_nm_gmpcare` is used for small
        problems (smaller than defined with
        :func:`~pymor.algorithms.lyapunov.mat_eqn_sparse_min_size`) and
        `lrnm` for large problems.

        Parameters
        ----------
        A
            The non-parametric |Operator| A.
        E
            The non-parametric |Operator| E or `None`.
        B
            The operator B as a |VectorArray| from `A.source`.
        C
            The operator C as a |VectorArray| from `A.source`.
        R
            The matrix R as a 2D |NumPy array| or `None`.
        trans
            Whether the first |Operator| in the Riccati equation is
            transposed.
        options
            The solver options to use (see
            :func:`ricc_lrcf_solver_options`).
        default_solver
            Default solver to use (pymess_lrnm,
            pymess_dense_nm_gmpcare).
            If `None`, chose solver depending on dimension `A`.

        Returns
        -------
        Z
            Low-rank Cholesky factor of the Riccati equation solution,
            |VectorArray| from `A.source`.
        """
        _solve_ricc_check_args(A, E, B, C, R, trans)
        if default_solver is None:
            default_solver = 'pymess_lrnm' if A.source.dim >= mat_eqn_sparse_min_size() else 'pymess_dense_nm_gmpcare'
        options = _parse_options(options, ricc_lrcf_solver_options(), default_solver, None, False)

        if options['type'] == 'pymess_dense_nm_gmpcare':
            X = _call_pymess_dense_nm_gmpare(A, E, B, C, R, trans=trans, options=options['opts'], plus=False,
                                             method_name='solve_ricc_lrcf')
            Z = _chol(X)
        elif options['type'] == 'pymess_lrnm':
            if R is not None:
                import scipy.linalg as spla
                Rc = spla.cholesky(R)                                 # R = Rc^T * Rc
                Rci = spla.solve_triangular(Rc, np.eye(Rc.shape[0]))  # R^{-1} = Rci * Rci^T
                if not trans:
                    C = C.lincomb(Rci.T)  # C <- Rci^T * C = (C^T * Rci)^T
                else:
                    B = B.lincomb(Rci.T)  # B <- B * Rci
            opts = options['opts']
            opts.type = pymess.MESS_OP_NONE if not trans else pymess.MESS_OP_TRANSPOSE
            eqn = RiccatiEquation(opts, A, E, B, C)
            Z, status = pymess.lrnm(eqn, opts)
            relres = status.res2_norm / status.res2_0
            if relres > opts.adi.res2_tol:
                logger = getLogger('pymor.bindings.pymess.solve_ricc_lrcf')
                logger.warning(f'Desired relative residual tolerance was not achieved '
                               f'({relres:e} > {opts.adi.res2_tol:e}).')
        else:
            raise ValueError(f'Unexpected Riccati equation solver ({options["type"]}).')

        return A.source.from_numpy(Z.T)

    def pos_ricc_lrcf_solver_options():
        """Return available positive Riccati solvers with default options for the pymess backend.

        Returns
        -------
        A dict of available solvers with default solver options.
        """
        return {'pymess_dense_nm_gmpcare': {'type': 'pymess_dense_nm_gmpcare',
                                            'opts': dense_nm_gmpcare_solver_options()}}

    def solve_pos_ricc_lrcf(A, E, B, C, R=None, trans=False, options=None):
        """Compute an approximate low-rank solution of a positive Riccati equation.

        See :func:`pymor.algorithms.riccati.solve_pos_ricc_lrcf` for a
        general description.

        This function uses `pymess.dense_nm_gmpcare`.

        Parameters
        ----------
        A
            The non-parametric |Operator| A.
        E
            The non-parametric |Operator| E or `None`.
        B
            The operator B as a |VectorArray| from `A.source`.
        C
            The operator C as a |VectorArray| from `A.source`.
        R
            The matrix R as a 2D |NumPy array| or `None`.
        trans
            Whether the first |Operator| in the Riccati equation is
            transposed.
        options
            The solver options to use (see
            :func:`pos_ricc_lrcf_solver_options`).

        Returns
        -------
        Z
            Low-rank Cholesky factor of the Riccati equation solution,
            |VectorArray| from `A.source`.
        """
        _solve_ricc_check_args(A, E, B, C, R, trans)
        options = _parse_options(options, pos_ricc_lrcf_solver_options(), 'pymess_dense_nm_gmpcare', None, False)

        if options['type'] == 'pymess_dense_nm_gmpcare':
            X = _call_pymess_dense_nm_gmpare(A, E, B, C, R, trans=trans, options=options['opts'], plus=True,
                                             method_name='solve_pos_ricc_lrcf')
            Z = _chol(X)
        else:
            raise ValueError(f'Unexpected positive Riccati equation solver ({options["type"]}).')

        return A.source.from_numpy(Z.T)

    def _call_pymess_dense_nm_gmpare(A, E, B, C, R, trans=False, options=None, plus=False, method_name=''):
        """Return the solution from pymess.dense_nm_gmpare solver."""
        A = to_matrix(A, format='dense')
        E = to_matrix(E, format='dense') if E else None
        B = B.to_numpy().T
        C = C.to_numpy()

        Q = B.dot(B.T) if not trans else C.T.dot(C)
        pymess_trans = pymess.MESS_OP_NONE if not trans else pymess.MESS_OP_TRANSPOSE
        if not trans:
            RinvC = spla.solve(R, C) if R is not None else C
            G = C.T.dot(RinvC)
        else:
            RinvBT = spla.solve(R, B.T) if R is not None else B.T
            G = B.dot(RinvBT)
        X, absres, relres = pymess.dense_nm_gmpare(None,
                                                   A, E, Q, G,
                                                   plus=plus, trans=pymess_trans,
                                                   linesearch=options['linesearch'],
                                                   maxit=options['maxit'],
                                                   absres_tol=options['absres_tol'],
                                                   relres_tol=options['relres_tol'],
                                                   nrm=options['nrm'])
        if absres > options['absres_tol']:
            logger = getLogger('pymor.bindings.pymess.' + method_name)
            logger.warning(f'Desired absolute residual tolerance was not achieved '
                           f'({absres:e} > {options["absres_tol"]:e}).')
        if relres > options['relres_tol']:
            logger = getLogger('pymor.bindings.pymess.' + method_name)
            logger.warning(f'Desired relative residual tolerance was not achieved '
                           f'({relres:e} > {options["relres_tol"]:e}).')

        return X

    class LyapunovEquation(pymess.Equation):
        """Lyapunov equation class for pymess

        Represents a (generalized) continuous-time algebraic Lyapunov
        equation:

        - if opt.type is `pymess.MESS_OP_NONE` and E is `None`:

          .. math::
              A X + X A^T + B B^T = 0,

        - if opt.type is `pymess.MESS_OP_NONE` and E is not `None`:

          .. math::
              A X E^T + E X A^T + B B^T = 0,

        - if opt.type is `pymess.MESS_OP_TRANSPOSE` and E is `None`:

          .. math::
              A^T X + X A + B^T B = 0,

        - if opt.type is `pymess.MESS_OP_TRANSPOSE` and E is not `None`:

          .. math::
              A^T X E + E^T X A + B^T B = 0.

        Parameters
        ----------
        opt
            pymess Options structure.
        A
            The non-parametric |Operator| A.
        E
            The non-parametric |Operator| E or `None`.
        B
            The operator B as a |VectorArray| from `A.source`.
        """

        def __init__(self, opt, A, E, B):
            super().__init__(name='LyapunovEquation', opt=opt, dim=A.source.dim)
            self.a = A
            self.e = E
            self.rhs = B.to_numpy().T
            self.p = []

        def ax_apply(self, op, y):
            y = self.a.source.from_numpy(y.T)
            if op == pymess.MESS_OP_NONE:
                x = self.a.apply(y)
            else:
                x = self.a.apply_adjoint(y)
            return x.to_numpy().T

        def ex_apply(self, op, y):
            if self.e is None:
                return y

            y = self.a.source.from_numpy(y.T)
            if op == pymess.MESS_OP_NONE:
                x = self.e.apply(y)
            else:
                x = self.e.apply_adjoint(y)
            return x.to_numpy().T

        def ainv_apply(self, op, y):
            y = self.a.source.from_numpy(y.T)
            if op == pymess.MESS_OP_NONE:
                x = self.a.apply_inverse(y)
            else:
                x = self.a.apply_inverse_adjoint(y)
            return x.to_numpy().T

        def einv_apply(self, op, y):
            if self.e is None:
                return y

            y = self.a.source.from_numpy(y.T)
            if op == pymess.MESS_OP_NONE:
                x = self.e.apply_inverse(y)
            else:
                x = self.e.apply_inverse_adjoint(y)
            return x.to_numpy().T

        def apex_apply(self, op, p, idx_p, y):
            y = self.a.source.from_numpy(y.T)
            if op == pymess.MESS_OP_NONE:
                x = self.a.apply(y)
                if self.e is None:
                    x += p * y
                else:
                    x += p * self.e.apply(y)
            else:
                x = self.a.apply_adjoint(y)
                if self.e is None:
                    x += p.conjugate() * y
                else:
                    x += p.conjugate() * self.e.apply_adjoint(y)
            return x.to_numpy().T

        def apeinv_apply(self, op, p, idx_p, y):
            y = self.a.source.from_numpy(y.T)
            e = IdentityOperator(self.a.source) if self.e is None else self.e

            if p.imag == 0:
                ape = self.a + p.real * e
            else:
                ape = self.a + p * e

            if op == pymess.MESS_OP_NONE:
                x = ape.apply_inverse(y)
            else:
                x = ape.apply_inverse_adjoint(y.conj()).conj()
            return x.to_numpy().T

        def parameter(self, arp_p, arp_m, B=None, K=None):
            return None

    class RiccatiEquation(pymess.Equation):
        """Riccati equation class for pymess

        Represents a Riccati equation

        - if opt.type is `pymess.MESS_OP_NONE` and E is `None`:

          .. math::
              A X + X A^T - X C^T C X + B B^T = 0,

        - if opt.type is `pymess.MESS_OP_NONE` and E is not `None`:

          .. math::
              A X E^T + E X A^T - E X C^T C X E^T + B B^T = 0,

        - if opt.type is `pymess.MESS_OP_TRANSPOSE` and E is `None`:

          .. math::
              A^T X + X A - X B B^T X + C^T C = 0,

        - if opt.type is `pymess.MESS_OP_TRANSPOSE` and E is not `None`:

          .. math::
              A^T X E + E^T X A - E X B B^T X E^T + C^T C = 0.

        Parameters
        ----------
        opt
            pymess Options structure.
        A
            The non-parametric |Operator| A.
        E
            The non-parametric |Operator| E or `None`.
        B
            The operator B as a |VectorArray| from `A.source`.
        C
            The operator C as a |VectorArray| from `A.source`.
        """

        def __init__(self, opt, A, E, B, C):
            super().__init__(name='RiccatiEquation', opt=opt, dim=A.source.dim)
            self.a = A
            self.e = E
            self.b = B.to_numpy().T
            self.c = C.to_numpy()
            self.rhs = self.b if opt.type == pymess.MESS_OP_NONE else self.c.T
            self.p = []

        def ax_apply(self, op, y):
            y = self.a.source.from_numpy(y.T)
            if op == pymess.MESS_OP_NONE:
                x = self.a.apply(y)
            else:
                x = self.a.apply_adjoint(y)
            return x.to_numpy().T

        def ex_apply(self, op, y):
            if self.e is None:
                return y

            y = self.a.source.from_numpy(y.T)
            if op == pymess.MESS_OP_NONE:
                x = self.e.apply(y)
            else:
                x = self.e.apply_adjoint(y)
            return x.to_numpy().T

        def ainv_apply(self, op, y):
            y = self.a.source.from_numpy(y.T)
            if op == pymess.MESS_OP_NONE:
                x = self.a.apply_inverse(y)
            else:
                x = self.a.apply_inverse_adjoint(y)
            return x.to_numpy().T

        def einv_apply(self, op, y):
            if self.e is None:
                return y

            y = self.a.source.from_numpy(y.T)
            if op == pymess.MESS_OP_NONE:
                x = self.e.apply_inverse(y)
            else:
                x = self.e.apply_inverse_adjoint(y)
            return x.to_numpy().T

        def apex_apply(self, op, p, idx_p, y):
            y = self.a.source.from_numpy(y.T)
            if op == pymess.MESS_OP_NONE:
                x = self.a.apply(y)
                if self.e is None:
                    x += p * y
                else:
                    x += p * self.e.apply(y)
            else:
                x = self.a.apply_adjoint(y)
                if self.e is None:
                    x += p.conjugate() * y
                else:
                    x += p.conjugate() * self.e.apply_adjoint(y)
            return x.to_numpy().T

        def apeinv_apply(self, op, p, idx_p, y):
            y = self.a.source.from_numpy(y.T)
            e = IdentityOperator(self.a.source) if self.e is None else self.e

            if p.imag == 0:
                ape = self.a + p.real * e
            else:
                ape = self.a + p * e

            if op == pymess.MESS_OP_NONE:
                x = ape.apply_inverse(y)
            else:
                x = ape.apply_inverse_adjoint(y.conj()).conj()
            return x.to_numpy().T

        def parameter(self, arp_p, arp_m, B=None, K=None):
            return None



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

from pymor.core.base import abstractmethod
from pymor.operators.interface import Operator
from pymor.operators.numpy import NumpyMatrixOperator
from pymor.vectorarrays.list import NumpyListVectorSpace


class ListVectorArrayOperatorBase(Operator):

    def _prepare_apply(self, U, mu, kind, least_squares=False):
        pass

    @abstractmethod
    def _apply_one_vector(self, u, mu=None, prepare_data=None):
        pass

    def _apply_inverse_one_vector(self, v, mu=None, initial_guess=None, least_squares=False, prepare_data=None):
        raise NotImplementedError

    def _apply_adjoint_one_vector(self, v, mu=None, prepare_data=None):
        raise NotImplementedError

    def _apply_inverse_adjoint_one_vector(self, u, mu=None, initial_guess=None, least_squares=False, prepare_data=None):
        raise NotImplementedError

    def apply(self, U, mu=None):
        assert U in self.source
        data = self._prepare_apply(U, mu, 'apply')
        V = [self._apply_one_vector(u, mu=mu, prepare_data=data) for u in U._list]
        return self.range.make_array(V)

    def apply_inverse(self, V, mu=None, initial_guess=None, least_squares=False):
        assert V in self.range
        assert initial_guess is None or initial_guess in self.source and len(initial_guess) == len(V)
        try:
            data = self._prepare_apply(V, mu, 'apply_inverse', least_squares=least_squares)
            U = [self._apply_inverse_one_vector(v, mu=mu,
                                                initial_guess=(initial_guess._list[i]
                                                               if initial_guess is not None else None),
                                                least_squares=least_squares, prepare_data=data)
                 for i, v in enumerate(V._list)]
        except NotImplementedError:
            return super().apply_inverse(V, mu=mu, least_squares=least_squares)
        return self.source.make_array(U)

    def apply_adjoint(self, V, mu=None):
        assert V in self.range
        try:
            data = self._prepare_apply(V, mu, 'apply_adjoint')
            U = [self._apply_adjoint_one_vector(v, mu=mu, prepare_data=data) for v in V._list]
        except NotImplementedError:
            return super().apply_adjoint(V, mu=mu)
        return self.source.make_array(U)

    def apply_inverse_adjoint(self, U, mu=None, initial_guess=None, least_squares=False):
        assert U in self.source
        try:
            data = self._prepare_apply(U, mu, 'apply_inverse_adjoint', least_squares=least_squares)
            V = [self._apply_inverse_adjoint_one_vector(u, mu=mu,
                                                        initial_guess=(initial_guess._list[i]
                                                                       if initial_guess is not None else None),
                                                        least_squares=least_squares, prepare_data=data)
                 for i, u in enumerate(U._list)]
        except NotImplementedError:
            return super().apply_inverse_adjoint(U, mu=mu, least_squares=least_squares)
        return self.range.make_array(V)


class LinearComplexifiedListVectorArrayOperatorBase(ListVectorArrayOperatorBase):

    linear = True

    @abstractmethod
    def _real_apply_one_vector(self, u, mu=None, prepare_data=None):
        pass

    def _real_apply_inverse_one_vector(self, v, mu=None, initial_guess=None, least_squares=False, prepare_data=None):
        raise NotImplementedError

    def _real_apply_adjoint_one_vector(self, v, mu=None, prepare_data=None):
        raise NotImplementedError

    def _real_apply_inverse_adjoint_one_vector(self, u, mu=None, initial_guess=None, least_squares=False,
                                               prepare_data=None):
        raise NotImplementedError

    def _apply_one_vector(self, u, mu=None, prepare_data=None):
        real_part = self._real_apply_one_vector(u.real_part, mu=mu, prepare_data=prepare_data)
        if u.imag_part is not None:
            imag_part = self._real_apply_one_vector(u.imag_part, mu=mu, prepare_data=prepare_data)
        else:
            imag_part = None
        return self.range.vector_type(real_part, imag_part)

    def _apply_inverse_one_vector(self, v, mu=None, initial_guess=None, least_squares=False, prepare_data=None):
        real_part = self._real_apply_inverse_one_vector(v.real_part, mu=mu,
                                                        initial_guess=(initial_guess.real_part
                                                                       if initial_guess is not None else None),
                                                        least_squares=least_squares,
                                                        prepare_data=prepare_data)
        if v.imag_part is not None:
            imag_part = self._real_apply_inverse_one_vector(v.imag_part, mu=mu,
                                                            initial_guess=(initial_guess.imag_part
                                                                           if initial_guess is not None else None),
                                                            least_squares=least_squares,
                                                            prepare_data=prepare_data)
        else:
            imag_part = None
        return self.source.vector_type(real_part, imag_part)

    def _apply_adjoint_one_vector(self, v, mu=None, prepare_data=None):
        real_part = self._real_apply_adjoint_one_vector(v.real_part, mu=mu, prepare_data=prepare_data)
        if v.imag_part is not None:
            imag_part = self._real_apply_adjoint_one_vector(v.imag_part, mu=mu, prepare_data=prepare_data)
        else:
            imag_part = None
        return self.source.vector_type(real_part, imag_part)

    def _apply_inverse_adjoint_one_vector(self, u, mu=None, initial_guess=None, least_squares=False, prepare_data=None):
        real_part = self._real_apply_inverse_adjoint_one_vector(u.real_part, mu=mu,
                                                                initial_guess=(initial_guess.real_part
                                                                               if initial_guess is not None else None),
                                                                least_squares=least_squares,
                                                                prepare_data=prepare_data)
        if u.imag_part is not None:
            imag_part = self._real_apply_inverse_adjoint_one_vector(u.imag_part, mu=mu,
                                                                    initial_guess=(initial_guess.imag_part
                                                                                   if initial_guess is not None
                                                                                   else None),
                                                                    least_squares=least_squares,
                                                                    prepare_data=prepare_data)
        else:
            imag_part = None
        return self.range.vector_type(real_part, imag_part)


class NumpyListVectorArrayMatrixOperator(ListVectorArrayOperatorBase, NumpyMatrixOperator):
    """Variant of |NumpyMatrixOperator| using |ListVectorArray| instead of |NumpyVectorArray|.

    This class is mainly intended for performance tests of |ListVectorArray|.
    In general |NumpyMatrixOperator| should be used instead of this class.

    Parameters
    ----------
    matrix
        The |NumPy array| which is to be wrapped.
    source_id
        The id of the operator's `source` |VectorSpace|.
    range_id
        The id of the operator's `range` |VectorSpace|.
    solver_options
        The |solver_options| for the operator.
    name
        Name of the operator.
    """

    def __init__(self, matrix, source_id=None, range_id=None, solver_options=None, name=None):
        super().__init__(matrix, source_id=source_id, range_id=range_id, solver_options=solver_options, name=name)
        self.source = NumpyListVectorSpace(matrix.shape[1], source_id)
        self.range = NumpyListVectorSpace(matrix.shape[0], range_id)

    def _apply_one_vector(self, u, mu=None, prepare_data=None):
        return self.matrix.dot(u._array)

    def _apply_inverse_one_vector(self, v, mu=None, initial_guess=None, least_squares=False, prepare_data=None):
        op = self.with_(new_type=NumpyMatrixOperator)
        u = op.apply_inverse(
            op.range.make_array(v._array),
            initial_guess=op.source.make_array(initial_guess._array) if initial_guess is not None else None,
            least_squares=least_squares
        ).to_numpy().ravel()

        return u

    def apply_adjoint(self, V, mu=None):
        return NumpyMatrixOperator.apply_adjoint(self, V, mu=mu)

    def apply_inverse_adjoint(self, U, mu=None, initial_guess=None, least_squares=False):
        return NumpyMatrixOperator.apply_inverse_adjoint(self, U, mu=mu,
                                                         initial_guess=initial_guess, least_squares=least_squares)

    def _assemble_lincomb(self, operators, coefficients, identity_shift=0., solver_options=None, name=None):
        lincomb = super()._assemble_lincomb(operators, coefficients, identity_shift)
        if lincomb is None:
            return None
        else:
            return lincomb.with_(new_type=NumpyListVectorArrayMatrixOperator)



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

"""This module contains a base class for implementing WorkerPool."""

import weakref

from pymor.core.base import ImmutableObject
from pymor.parallel.interface import WorkerPool, RemoteObject


class WorkerPoolDefaultImplementations:

    def scatter_array(self, U, copy=True):
        slice_len = len(U) // len(self) + (1 if len(U) % len(self) else 0)
        if copy:
            slices = []
            for i in range(len(self)):
                slices.append(U[i*slice_len:min((i+1)*slice_len, len(U))].copy())
        else:
            slices = [U.empty() for _ in range(len(self))]
            for s in slices:
                s.append(U[:min(slice_len, len(U))], remove_from_other=True)
        remote_U = self.push(U.empty())
        del U
        self.map(_append_array_slice, slices, U=remote_U)
        return remote_U

    def scatter_list(self, l):
        slice_len = len(l) // len(self) + (1 if len(l) % len(self) else 0)
        slices = []
        for i in range(len(self)):
            slices.append(l[i*slice_len:(i+1)*slice_len])
        del l
        remote_l = self.push([])
        self.map(_append_list_slice, slices, l=remote_l)
        return remote_l


class WorkerPoolBase(WorkerPoolDefaultImplementations, WorkerPool):

    def __init__(self):
        self._pushed_immutable_objects = {}

    def push(self, obj):
        if isinstance(obj, ImmutableObject):
            uid = obj.uid
            if uid not in self._pushed_immutable_objects:
                remote_id = self._push_object(obj)
                self._pushed_immutable_objects[uid] = (remote_id, 1)
            else:
                remote_id, ref_count = self._pushed_immutable_objects[uid]
                self._pushed_immutable_objects[uid] = (remote_id, ref_count + 1)
            return GenericRemoteObject(self, remote_id, uid=uid)
        else:
            remote_id = self._push_object(obj)
            return GenericRemoteObject(self, remote_id)

    def _map_kwargs(self, kwargs):
        pushed_immutable_objects = self._pushed_immutable_objects
        return {k: (pushed_immutable_objects.get(v.uid, (v, 0))[0] if isinstance(v, ImmutableObject) else
                    v.remote_id if isinstance(v, RemoteObject) else
                    v)
                for k, v in kwargs.items()}

    def apply(self, function, *args, **kwargs):
        kwargs = self._map_kwargs(kwargs)
        return self._apply(function, *args, **kwargs)

    def apply_only(self, function, worker, *args, **kwargs):
        kwargs = self._map_kwargs(kwargs)
        return self._apply_only(function, worker, *args, **kwargs)

    def map(self, function, *args, **kwargs):
        kwargs = self._map_kwargs(kwargs)
        chunks = self._split_into_chunks(len(self), *args)
        return self._map(function, chunks, **kwargs)

    def _split_into_chunks(self, count, *args):
        lens = list(map(len, args))
        min_len = min(lens)
        max_len = max(lens)
        assert min_len == max_len
        chunk_size = max_len // count + (1 if max_len % count > 0 else 0)

        def split_arg(arg):
            for _ in range(count):
                chunk, arg = arg[:chunk_size], arg[chunk_size:]
                yield chunk
        from itertools import chain
        for arg in args:
            assert list(chain(*split_arg(arg))) == arg
        chunks = tuple(list(split_arg(arg)) for arg in args)
        return chunks


class GenericRemoteObject(RemoteObject):

    def __init__(self, pool, remote_id, uid=None):
        self.pool = weakref.ref(pool)
        self.remote_id = remote_id
        self.uid = uid

    def _remove(self):
        pool = self.pool()
        if self.uid is not None:
            remote_id, ref_count = pool._pushed_immutable_objects.pop(self.uid)
            if ref_count > 1:
                pool._pushed_immutable_objects[self.remote_id] = (remote_id, ref_count - 1)
            else:
                pool._remove_object(remote_id)
        else:
            pool._remove_object(self.remote_id)


def _append_array_slice(s, U=None):
    U.append(s, remove_from_other=True)


def _append_list_slice(s, l=None):
    l.extend(s)
